import pytest
from tqdm import tqdm
from transformers import AutoTokenizer

from oumi.core.registry import REGISTRY

USE_STREAMING: bool = True
LIMIT_SAMPLES: int = 1_000_000  # set to 0 to iterate through the entire dataset


@pytest.fixture(
    # (dataset_name, dataset_subset)
    params=[
        ("allenai/c4", "en"),
        ("allenai/dolma", "v1_7"),
        ("bigcode/starcoderdata", "python"),
        ("bigcode/the-stack", "python"),
        ("cerebras/SlimPajama-627B", None),
        ("EleutherAI/pile", "default"),
        ("HuggingFaceFW/fineweb-edu", "sample-10BT"),
        ("nampdn-ai/tiny-textbooks", None),
        ("PleIAs/YouTube-Commons", None),
        ("roneneldan/TinyStories", None),
        ("Salesforce/wikitext", "wikitext-2-raw-v1"),
        ("tiiuae/falcon-refinedweb", None),
        ("togethercomputer/RedPajama-Data-1T", "common_crawl"),
        ("togethercomputer/RedPajama-Data-V2", "default"),
        ("wikimedia/wikipedia", "20231101.en"),
    ]
)
def dataset_fixture(request):
    dataset_name, dataset_subset = request.param
    dataset_class = REGISTRY.get_dataset(dataset_name)
    if dataset_class is None:
        pytest.fail(f"Dataset {dataset_name} not found in registry")

    tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=True)
    tokenizer.eos_token = "</s>"

    return dataset_name, dataset_class(
        dataset_name=dataset_name,
        subset=dataset_subset,
        split="train",
        tokenizer=tokenizer,
        seq_length=64,
        stream=USE_STREAMING,
    )


@pytest.mark.skip(
    reason="This test is very time consuming, and should be run manually."
)
def test_dataset_iteration(dataset_fixture):
    dataset_name, dataset = dataset_fixture
    assert dataset is not None

    # Iterate through all the items in the dataset
    for batch_idx, batch in enumerate(tqdm(dataset)):
        assert isinstance(batch, dict), (
            f"Invalid batch format for dataset {dataset_name} at batch {batch_idx}"
        )
        assert "input_ids" in batch, (
            f"Missing 'input_ids' in batch {batch_idx} for dataset {dataset_name}"
        )
        assert "attention_mask" in batch, (
            f"Missing 'attention_mask' in batch {batch_idx} for dataset {dataset_name}"
        )

        if LIMIT_SAMPLES > 0 and batch_idx >= LIMIT_SAMPLES:
            break



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("bigcode/the-stack")
class TheStackDataset(BasePretrainingDataset):
    """A dataset containing over 6TB of permissively-licensed source code files.

    The Stack was created as part of the BigCode Project, an open scientific
    collaboration working on the responsible development of Large Language Models
    for Code (Code LLMs). It serves as a pre-training dataset for Code LLMs,
    enabling the synthesis of programs from natural language descriptions and
    other code snippets, and covers 358 programming languages.

    The dataset contains code in multiple natural languages, primarily found in
    comments and docstrings. It supports tasks such as code completion,
    documentation generation, and auto-completion of code snippets.

    See Also:
        - Huggingface hub: https://huggingface.co/datasets/bigcode/the-stack
        - Homepage: https://www.bigcode-project.org/
        - Repository: https://github.com/bigcode-project
        - Paper: https://arxiv.org/abs/2211.15533

    Data Fields:
        - content (string): The content of the file.
        - size (integer): Size of the uncompressed file.
        - lang (string): The programming language.
        - ext (string): File extension.
        - avg_line_length (float): The average line-length of the file.
        - max_line_length (integer): The maximum line-length of the file.
        - alphanum_fraction (float): The fraction of alphanumeric characters.
        - hexsha (string): Unique git hash of file.
        - max_{stars|forks|issues}_repo_path (string): Path to file in repo.
        - max_{stars|forks|issues}_repo_name (string): Name of repo.
        - max_{stars|forks|issues}_repo_head_hexsha (string): Hexsha of repo head.
        - max_{stars|forks|issues}_repo_licenses (string): Licenses in repository.
        - max_{stars|forks|issues}_count (integer): Number of stars/forks/issues.
        - max_{stars|forks|issues}_repo_{stars|forks|issues}_min_datetime (string):
          First timestamp of a stars/forks/issues event.
        - max_{stars|forks|issues}_repo_{stars|forks|issues}_max_datetime (string):
          Last timestamp of a stars/forks/issues event.
    """

    default_dataset = "bigcode/the-stack"
    _data_column = "content"



import json
import tempfile
from pathlib import Path

import jsonlines
import pandas as pd
import pytest

from oumi.core.types.conversation import Conversation
from oumi.datasets import TextSftJsonLinesDataset


@pytest.fixture
def sample_jsonlines_data():
    return [
        {
            "messages": [
                {"role": "user", "content": "Hello, how are you?"},
                {
                    "role": "assistant",
                    "content": "I'm doing well, thank you! How can I assist you today?",
                },
                {
                    "role": "user",
                    "content": "Can you explain what machine learning is?",
                },
                {
                    "role": "assistant",
                    "content": "Certainly! Machine learning is a"
                    " branch of artificial intelligence...",
                },
            ]
        }
    ]


@pytest.fixture
def sample_oumi_data():
    return [
        {
            "messages": [
                {
                    "role": "user",
                    "content": "Translate the following English text to French\n\n"
                    "Hello, how are you?",
                },
                {
                    "role": "assistant",
                    "content": "Bonjour, comment allez-vous ?",
                },
            ]
        }
    ]


@pytest.fixture
def sample_alpaca_data():
    return [
        {
            "instruction": "Translate the following English text to French",
            "input": "Hello, how are you?",
            "output": "Bonjour, comment allez-vous ?",
        }
    ]


@pytest.fixture
def sample_conversations_data():
    """Sample data for conversations format - conversation key with conversation json"""
    return [
        {
            "conversation": {
                "conversation_id": "123",
                "messages": [
                    {"role": "user", "content": "What is the capital of France?"},
                    {"role": "assistant", "content": "The capital of France is Paris."},
                ],
                "metadata": {"key": "value"},
            }
        }
    ]


def test_text_jsonlines_init_with_data(sample_jsonlines_data):
    dataset = TextSftJsonLinesDataset(data=sample_jsonlines_data)
    assert len(dataset._data) == 1
    assert ["_messages_column"] == dataset._data.columns

    conversation = dataset.conversation(0)

    assert isinstance(conversation, Conversation)
    assert len(conversation.messages) == 4
    assert conversation.messages[0].role == "user"
    assert conversation.messages[0].content == "Hello, how are you?"
    assert conversation.messages[1].role == "assistant"
    assert (
        conversation.messages[1].content
        == "I'm doing well, thank you! How can I assist you today?"
    )


def test_text_jsonlines_init_with_dataset_path(sample_jsonlines_data):
    with tempfile.TemporaryDirectory() as folder:
        valid_jsonlines_filename = Path(folder) / "valid_path.jsonl"
        with jsonlines.open(valid_jsonlines_filename, mode="w") as writer:
            writer.write_all(sample_jsonlines_data)

        dataset = TextSftJsonLinesDataset(dataset_path=valid_jsonlines_filename)
        assert len(dataset._data) == 1
        assert ["_messages_column"] == dataset._data.columns

        conversation = dataset.conversation(0)

        assert isinstance(conversation, Conversation)
        assert len(conversation.messages) == 4
        assert conversation.messages[0].role == "user"
        assert conversation.messages[0].content == "Hello, how are you?"
        assert conversation.messages[1].role == "assistant"
        assert (
            conversation.messages[1].content
            == "I'm doing well, thank you! How can I assist you today?"
        )


def test_text_jsonlines_init_with_invalid_input(sample_jsonlines_data):
    with tempfile.TemporaryDirectory() as folder:
        valid_jsonlines_filename = Path(folder) / "valid_path.jsonl"
        with jsonlines.open(valid_jsonlines_filename, mode="w") as writer:
            writer.write_all(sample_jsonlines_data)

        with pytest.raises(ValueError, match="Dataset path or data must be provided"):
            TextSftJsonLinesDataset()

        with pytest.raises(
            FileNotFoundError,
            match="Provided path does not exist: 'invalid_path.jsonl'.",
        ):
            TextSftJsonLinesDataset(dataset_path="invalid_path.jsonl")

        with pytest.raises(
            ValueError,
            match="Either dataset_path or data must be provided, but not both",
        ):
            TextSftJsonLinesDataset(dataset_path=valid_jsonlines_filename, data=[])

        # Directory ending with .jsonl
        temp_dir_name = Path(folder) / "subdir.jsonl"
        temp_dir_name.mkdir()
        with pytest.raises(
            ValueError,
            match="Provided path is a directory, expected a file",
        ):
            TextSftJsonLinesDataset(dataset_path=temp_dir_name)


def test_text_jsonlines_load_data():
    dataset = TextSftJsonLinesDataset(
        data=[{"messages": [{"role": "user", "content": "Test"}]}]
    )
    loaded_data = dataset._load_data()
    assert isinstance(loaded_data, pd.DataFrame)
    assert len(loaded_data) == 1
    assert ["_messages_column"] == loaded_data.columns


def test_text_jsonlines_getitem():
    data = [
        {
            "messages": [
                {"role": "user", "content": "Hello"},
                {"role": "assistant", "content": "Hi there!"},
            ]
        },
        {
            "messages": [
                {"role": "user", "content": "How are you?"},
                {"role": "assistant", "content": "I'm good, thanks!"},
            ]
        },
    ]
    dataset = TextSftJsonLinesDataset(data=data)

    item = dataset.conversation(0)
    assert len(item.messages) == 2

    with pytest.raises(IndexError):
        _ = dataset.conversation(2)


def test_text_jsonlines_len():
    data = [
        {"messages": [{"role": "user", "content": "Hello"}]},
        {"messages": [{"role": "user", "content": "How are you?"}]},
    ]
    dataset = TextSftJsonLinesDataset(data=data)
    assert len(dataset) == 2


def test_load_oumi_format_jsonl(sample_oumi_data):
    with tempfile.TemporaryDirectory() as folder:
        file_path = Path(folder) / "oumi_data.jsonl"
        with jsonlines.open(file_path, mode="w") as writer:
            writer.write_all(sample_oumi_data)

        dataset = TextSftJsonLinesDataset(dataset_path=file_path)
        assert dataset._format == "oumi"
        assert len(dataset) == 1

        conversation = dataset.conversation(0)
        assert isinstance(conversation, Conversation)
        assert len(conversation.messages) == 2
        assert conversation.messages[0].role == "user"
        assert (
            conversation.messages[0].content
            == "Translate the following English text to French\n\nHello, how are you?"
        )
        assert conversation.messages[1].role == "assistant"
        assert conversation.messages[1].content == "Bonjour, comment allez-vous ?"


def test_load_alpaca_format_json(sample_alpaca_data):
    with tempfile.TemporaryDirectory() as folder:
        file_path = Path(folder) / "alpaca_data.json"
        with open(file_path, "w") as f:
            json.dump(sample_alpaca_data, f)

        dataset = TextSftJsonLinesDataset(dataset_path=file_path, format="alpaca")
        assert dataset._format == "alpaca"
        assert len(dataset) == 1

        conversation = dataset.conversation(0)
        assert isinstance(conversation, Conversation)
        assert len(conversation.messages) == 2
        assert conversation.messages[0].role == "user"
        assert (
            conversation.messages[0].content
            == "Translate the following English text to French\n\nHello, how are you?"
        )
        assert conversation.messages[1].role == "assistant"
        assert conversation.messages[1].content == "Bonjour, comment allez-vous ?"


def test_auto_detect_format(sample_oumi_data, sample_alpaca_data):
    oumi_dataset = TextSftJsonLinesDataset(data=sample_oumi_data)
    assert oumi_dataset._format == "oumi"

    alpaca_dataset = TextSftJsonLinesDataset(data=sample_alpaca_data)
    assert alpaca_dataset._format == "alpaca"


def test_conversations_format(sample_conversations_data):
    """Test loading and processing conversations format data."""
    dataset = TextSftJsonLinesDataset(data=sample_conversations_data)
    assert dataset._format == "conversations"
    assert len(dataset) == 1

    conversation = dataset.conversation(0)
    assert isinstance(conversation, Conversation)
    assert len(conversation.messages) == 2
    assert conversation.messages[0].role == "user"
    assert conversation.messages[0].content == "What is the capital of France?"
    assert conversation.messages[1].role == "assistant"
    assert conversation.messages[1].content == "The capital of France is Paris."

    assert conversation.conversation_id == "123"
    assert conversation.metadata == {"key": "value"}


def test_invalid_format():
    with pytest.raises(ValueError, match="Invalid format:"):
        TextSftJsonLinesDataset(data=[{"messages": []}], format="invalid_format")


def test_unsupported_file_extension():
    with tempfile.TemporaryDirectory() as folder:
        file_path = Path(folder) / "data.txt"
        file_path.touch()

        with pytest.raises(ValueError, match="Unsupported file format:"):
            TextSftJsonLinesDataset(dataset_path=file_path)


def test_invalid_data_structure():
    with pytest.raises(ValueError, match="Invalid data format"):
        TextSftJsonLinesDataset(data=[["not a dict"]])  # type: ignore[arg-type]


def test_undetectable_format():
    with pytest.raises(ValueError, match="Unable to auto-detect format"):
        TextSftJsonLinesDataset(data=[{"unknown_key": "value"}])


def test_alpaca_to_conversation():
    alpaca_turn = {
        "instruction": "Translate to French",
        "input": "Hello",
        "output": "Bonjour",
    }
    dataset = TextSftJsonLinesDataset(data=[alpaca_turn], format="alpaca")
    conversation = dataset.conversation(0)

    assert isinstance(conversation, Conversation)
    assert len(conversation.messages) == 2
    assert conversation.messages[0].role == "user"
    assert conversation.messages[0].content == "Translate to French\n\nHello"
    assert conversation.messages[1].role == "assistant"
    assert conversation.messages[1].content == "Bonjour"


def test_alpaca_missing_keys():
    invalid_alpaca_turn = {
        "instruction": "Translate to French",
        "input": "Hello",
        # Missing "output" key
    }
    dataset = TextSftJsonLinesDataset(data=[invalid_alpaca_turn], format="alpaca")

    with pytest.raises(ValueError, match="Invalid Alpaca format"):
        dataset.conversation(0)


def test_oumi_and_alpaca_format_prompt_equality(sample_oumi_data, sample_alpaca_data):
    oumi_dataset = TextSftJsonLinesDataset(data=sample_oumi_data)
    alpaca_dataset = TextSftJsonLinesDataset(data=sample_alpaca_data)

    oumi_conversation = oumi_dataset.conversation(0)
    alpaca_conversation = alpaca_dataset.conversation(0)

    assert len(oumi_conversation.messages) == len(alpaca_conversation.messages)
    assert oumi_conversation.messages[0].role == alpaca_conversation.messages[0].role
    assert (
        oumi_conversation.messages[0].content == alpaca_conversation.messages[0].content
    )
    assert oumi_conversation.messages[1].role == alpaca_conversation.messages[1].role
    assert (
        oumi_conversation.messages[1].content == alpaca_conversation.messages[1].content
    )



import json
from unittest.mock import AsyncMock, patch

import pydantic
import pytest

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
    RemoteParams,
)
from oumi.core.configs.params.guided_decoding_params import GuidedDecodingParams
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.gemini_inference_engine import GoogleGeminiInferenceEngine


@pytest.fixture
def gemini_engine():
    model_params = ModelParams(model_name="gemini-model")
    return GoogleGeminiInferenceEngine(
        model_params,
        remote_params=RemoteParams(
            api_url="https://example.com/api",
            api_key="dummy_api_key",
            num_workers=1,
            max_retries=3,
            connection_timeout=30,
            politeness_policy=0.1,
        ),
    )


@pytest.fixture
def generation_params():
    return GenerationParams(
        max_new_tokens=100,
        temperature=0.7,
        top_p=0.9,
    )


@pytest.fixture
def inference_config(generation_params):
    return InferenceConfig(
        generation=generation_params,
        remote_params=RemoteParams(
            api_url="https://example.com/api",
            api_key="dummy_api_key",
            num_workers=1,
            max_retries=3,
            connection_timeout=30,
            politeness_policy=0.1,
        ),
    )


def test_gemini_convert_conversation(gemini_engine, generation_params):
    """Test basic conversation conversion without special features."""
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
            Message(content="Hi there!", role=Role.ASSISTANT),
            Message(content="How are you?", role=Role.USER),
        ]
    )

    api_input = gemini_engine._convert_conversation_to_api_input(
        conversation, generation_params, gemini_engine._model_params
    )

    assert api_input["model"] == "gemini-model"
    assert len(api_input["messages"]) == 3
    assert api_input["max_completion_tokens"] == 100
    assert api_input["temperature"] == 0.7
    assert api_input["top_p"] == 0.9
    assert api_input["n"] == 1


def test_gemini_convert_conversation_with_guided_decoding(
    gemini_engine, generation_params
):
    """Test conversation conversion with JSON schema guided decoding."""

    class TestSchema(pydantic.BaseModel):
        name: str
        age: int

    generation_params.guided_decoding = GuidedDecodingParams(json=TestSchema)
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    api_input = gemini_engine._convert_conversation_to_api_input(
        conversation, generation_params, gemini_engine._model_params
    )

    assert "response_format" in api_input
    assert api_input["response_format"]["type"] == "json_schema"
    assert api_input["response_format"]["json_schema"]["name"] == "TestSchema"
    assert "properties" in api_input["response_format"]["json_schema"]["schema"]


@pytest.mark.parametrize(
    "json_schema",
    [
        {"type": "object", "properties": {"name": {"type": "string"}}},  # dict
        '{"type": "object", "properties": {"name": {"type": "string"}}}',  # str
    ],
)
def test_gemini_convert_conversation_with_json_schema_variations(
    gemini_engine, generation_params, json_schema
):
    """Test conversation conversion with different JSON schema formats."""
    generation_params.guided_decoding = GuidedDecodingParams(json=json_schema)
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    api_input = gemini_engine._convert_conversation_to_api_input(
        conversation, generation_params, gemini_engine._model_params
    )

    assert "response_format" in api_input
    assert api_input["response_format"]["type"] == "json_schema"
    assert api_input["response_format"]["json_schema"]["name"] == "Response"


def test_gemini_convert_conversation_invalid_schema(gemini_engine, generation_params):
    """Test that invalid schema types raise appropriate errors."""
    generation_params.guided_decoding = GuidedDecodingParams(json=123)  # Invalid type
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    with pytest.raises(ValueError) as exc_info:
        gemini_engine._convert_conversation_to_api_input(
            conversation, generation_params, gemini_engine._model_params
        )

    assert "unsupported JSON schema type" in str(exc_info.value)


@pytest.mark.asyncio
async def test_gemini_infer_online(gemini_engine, inference_config):
    """Test online inference with Gemini."""
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    with patch.object(
        gemini_engine,
        "_infer",
        new_callable=AsyncMock,
        side_effect=lambda convs, config: convs,
    ):
        results = gemini_engine.infer([conversation], inference_config)

    assert len(results) == 1
    assert results[0].messages == conversation.messages
    assert results[0].metadata == conversation.metadata
    if conversation.conversation_id is not None:
        assert results[0].conversation_id == conversation.conversation_id


def test_gemini_infer_from_file(gemini_engine, inference_config, tmp_path):
    """Test file-based inference with Gemini."""
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    input_file = tmp_path / "input.jsonl"
    with open(input_file, "w") as f:
        json.dump(conversation.to_dict(), f)
        f.write("\n")

    with patch.object(
        gemini_engine,
        "_infer",
        new_callable=AsyncMock,
        side_effect=lambda convs, config: convs,
    ):
        inference_config.input_path = str(input_file)
        results = gemini_engine.infer(inference_config=inference_config)

    assert len(results) == 1
    assert results[0].messages == conversation.messages
    assert results[0].metadata == conversation.metadata
    if conversation.conversation_id is not None:
        assert results[0].conversation_id == conversation.conversation_id


def test_gemini_batch_prediction_disabled(gemini_engine, inference_config):
    conversation = Conversation(
        messages=[
            Message(content="Hello", role=Role.USER),
        ]
    )

    with pytest.raises(NotImplementedError):
        gemini_engine.infer_batch([conversation], inference_config)


def test_remote_params_defaults():
    gemini_engine = GoogleGeminiInferenceEngine(
        model_params=ModelParams(model_name="some_model"),
    )
    assert gemini_engine._remote_params.num_workers == 2
    assert gemini_engine._remote_params.politeness_policy == 60.0



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SafeDict(dict):
    def __init__(self, missing_values_allowed: bool, *args, **kwargs):
        """Initialize the SafeDict with the missing_values_allowed flag."""
        self.missing_values_allowed = missing_values_allowed
        self.placeholder_names = set()
        super().__init__(*args, **kwargs)

    def __missing__(self, key: str) -> str:
        """Handle missing keys in the dictionary."""
        self.placeholder_names.add(key)
        if self.missing_values_allowed:
            return "{" + key + "}"
        else:
            raise ValueError(f"Missing value for placeholder: {key}")


def resolve_placeholders(
    text: str,
    values_dict: dict[str, str],
    missing_values_allowed: bool = False,
) -> str:
    """Resolve placeholder {variables} in the provided text from the values_dict."""
    return text.format_map(SafeDict(missing_values_allowed, values_dict))


def get_placeholders(text: str) -> set[str]:
    """Extract placeholder variable names from text with {variable} syntax."""
    safe_dict = SafeDict(missing_values_allowed=True)
    text.format_map(safe_dict)
    return safe_dict.placeholder_names



from pathlib import Path

from oumi.datasets import TextSftJsonLinesDataset


def test_data_format_loading():
    """Tests demo examples are correctly loaded in both json and jsonl formats."""
    current_dir = Path(__file__).resolve().parent
    data_top_dir = current_dir / "../../../data/dataset_examples"

    for format in ["alpaca", "oumi"]:
        all_data = []
        for ending in ["json", "jsonl"]:
            current = TextSftJsonLinesDataset(
                dataset_path=data_top_dir / (format + "_format." + ending)
            )
            all_data.append(current)

        json_datum = all_data[0]
        jsonl_datum = all_data[1]
        for i in range(len(json_datum)):
            assert json_datum.conversation(i) == jsonl_datum.conversation(i), (
                "Data from json and jsonl files should be the same by construction."
            )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for sample analyzer plugins."""

from abc import ABC, abstractmethod
from typing import Optional

import pandas as pd


class SampleAnalyzer(ABC):
    """Base class for sample analyzer plugins that analyze individual samples.

    All analyzers work with pandas DataFrames for efficient processing.
    """

    @abstractmethod
    def analyze_sample(
        self,
        df: pd.DataFrame,
        schema: Optional[dict] = None,
    ) -> pd.DataFrame:
        """Analyze text fields and return analysis results.

        This method performs analysis on the input DataFrame and returns
        the DataFrame with added analysis columns. All analyzers must implement this
        method.

        Args:
            df: Input DataFrame with text fields
            schema: Column schema dict to identify text fields

        Returns:
            DataFrame with added analysis columns
        """
        pass



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from transformers.optimization import Adafactor

from oumi.core.configs import TrainingParams
from oumi.utils.torch_naming_heuristics import group_trainable_params


def build_optimizer(
    model: torch.nn.Module, config: TrainingParams
) -> torch.optim.Optimizer:
    """Builds and returns a PyTorch optimizer based on the provided configuration.

    See pytorch documentation for more information on available optimizers:
    https://pytorch.org/docs/stable/optim.html

    Args:
        model: The model whose parameters will be optimized.
        config: The configuration object containing optimizer parameters.

    Returns:
        Optimizer: The constructed PyTorch optimizer.
    """
    optimizer_name = config.optimizer.lower()

    # Get parameters that require optimization, grouped by weight decay.
    trainable_param_groups = group_trainable_params(model, config.weight_decay)

    fused_available = torch.cuda.is_available()

    if optimizer_name == "adam":
        return torch.optim.Adam(
            trainable_param_groups,
            lr=config.learning_rate,
            betas=(config.adam_beta1, config.adam_beta2),
            eps=config.adam_epsilon,
            fused=fused_available,
        )
    elif optimizer_name in ("adamw", "adamw_torch", "adamw_torch_fused"):
        return torch.optim.AdamW(
            trainable_param_groups,
            lr=config.learning_rate,
            betas=(config.adam_beta1, config.adam_beta2),
            eps=config.adam_epsilon,
            fused=fused_available,
        )
    elif optimizer_name in (
        "adamw_8bit",
        "paged_adamw_8bit",
        "paged_adamw",
        "paged_adamw_32bit",
    ):
        try:
            import bitsandbytes  # pyright: ignore[reportMissingImports]
        except ImportError:
            raise ImportError(
                "bitsandbytes is not installed. "
                "Please install it with `pip install bitsandbytes` "
                "to use 8-bit or paged optimizers."
            )

        if optimizer_name in ("adamw_8bit", "paged_adamw_8bit"):
            return bitsandbytes.optim.AdamW(
                trainable_param_groups,
                lr=config.learning_rate,
                betas=(config.adam_beta1, config.adam_beta2),
                eps=config.adam_epsilon,
                weight_decay=config.weight_decay,
                optim_bits=8,
                is_paged=optimizer_name == "paged_adamw_8bit",
            )
        else:  # paged_adamw or paged_adamw_32bit
            return bitsandbytes.optim.PagedAdamW(
                trainable_param_groups,
                lr=config.learning_rate,
                betas=(config.adam_beta1, config.adam_beta2),
                eps=config.adam_epsilon,
                weight_decay=config.weight_decay,
            )
    elif optimizer_name == "sgd":
        return torch.optim.SGD(
            trainable_param_groups,
            lr=config.learning_rate,
            momentum=config.sgd_momentum,
            fused=fused_available,
        )
    elif optimizer_name == "adafactor":
        return Adafactor(
            trainable_param_groups,
            lr=config.learning_rate,
            relative_step=False,
            scale_parameter=False,
        )
    else:
        raise ValueError(f"Unsupported optimizer: {optimizer_name}")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from oumi.core.configs.params.synthesis_params import DatasetSource
from oumi.core.synthesis.dataset_ingestion import (
    DatasetPath,
    DatasetReader,
    DatasetStorageType,
)


def test_invalid_path():
    """Test invalid path."""
    with pytest.raises(ValueError, match="Invalid path"):
        DatasetPath("invalid:path")


def test_local_path():
    """Test initialization with local JSONL path."""
    path = DatasetPath("path/to/data/file.jsonl")
    assert path.get_path_str() == "path/to/data/file.jsonl"
    assert path.get_storage_type() == DatasetStorageType.LOCAL

    path = DatasetPath("path/to/data/file.csv")
    assert path.get_path_str() == "path/to/data/file.csv"
    assert path.get_storage_type() == DatasetStorageType.LOCAL

    path = DatasetPath("path/to/data/file.tsv")
    assert path.get_path_str() == "path/to/data/file.tsv"
    assert path.get_storage_type() == DatasetStorageType.LOCAL

    path = DatasetPath("path/to/data/file.parquet")
    assert path.get_path_str() == "path/to/data/file.parquet"
    assert path.get_storage_type() == DatasetStorageType.LOCAL

    path = DatasetPath("path/to/data/file.json")
    assert path.get_path_str() == "path/to/data/file.json"
    assert path.get_storage_type() == DatasetStorageType.LOCAL

    path = DatasetPath("path/to/data/*.jsonl")
    assert path.get_path_str() == "path/to/data/*.jsonl"
    assert path.get_storage_type() == DatasetStorageType.LOCAL


def test_huggingface_path():
    """Test get_path_str for HuggingFace path."""
    path = DatasetPath("hf:repo_id/dataset_name")
    assert path.get_path_str() == "repo_id/dataset_name"
    assert path.get_storage_type() == DatasetStorageType.HF


def test_oumi_path():
    """Test get_path_str for Oumi path."""
    path = DatasetPath("oumi:123")
    assert path.get_path_str() == "123"
    assert path.get_storage_type() == DatasetStorageType.OUMI


def test_get_file_extension_non_local():
    """Test that get_file_extension raises error for non-local storage types."""
    hf_path = DatasetPath("hf:repo/dataset")
    with pytest.raises(
        NotImplementedError, match="No extension for DatasetStorageType.HF storage type"
    ):
        hf_path.get_file_extension()

    oumi_path = DatasetPath("oumi:123")
    with pytest.raises(
        NotImplementedError,
        match="No extension for DatasetStorageType.OUMI storage type",
    ):
        oumi_path.get_file_extension()


@pytest.fixture
def reader():
    """Create a DatasetReader instance."""
    return DatasetReader()


@pytest.fixture
def sample_data():
    """Sample data for testing."""
    return [
        {"id": 1, "name": "Alice", "age": 25},
        {"id": 2, "name": "Bob", "age": 30},
        {"id": 3, "name": "Charlie", "age": 35},
    ]


@pytest.fixture
def sample_dataframe(sample_data):
    """Sample pandas DataFrame for testing."""
    return pd.DataFrame(sample_data)


def test_read_from_huggingface(reader, sample_data):
    """Test reading from HuggingFace dataset."""
    data_source = DatasetSource(path="hf:test/dataset")

    with patch("oumi.core.synthesis.dataset_ingestion.load_dataset") as mock_load:
        mock_dataset = MagicMock()
        mock_dataset.__iter__ = MagicMock(return_value=iter(sample_data))
        mock_load.return_value = mock_dataset

        result = reader.read(data_source)

        mock_load.assert_called_once_with("test/dataset", split=None, revision=None)
        assert result == sample_data


def test_read_from_hf_with_parameters(reader, sample_data):
    """Test reading from HuggingFace with split and revision parameters."""
    with patch("oumi.core.synthesis.dataset_ingestion.load_dataset") as mock_load:
        mock_dataset = MagicMock()
        mock_dataset.__iter__ = MagicMock(return_value=iter(sample_data))
        mock_load.return_value = mock_dataset

        result = reader._read_from_hf("test/dataset", split="train", revision="v1.0")

        mock_load.assert_called_once_with(
            "test/dataset",
            split="train",
            revision="v1.0",
        )
        assert result == sample_data


def test_read_from_oumi_dataset(reader, sample_data):
    """Test reading from Oumi dataset."""
    data_source = DatasetSource(path="oumi:123")

    with patch("oumi.core.synthesis.dataset_ingestion.load_dataset") as mock_load:
        mock_dataset = MagicMock()
        mock_dataset.__iter__ = MagicMock(return_value=iter(sample_data))
        mock_load.return_value = mock_dataset

        with pytest.raises(
            NotImplementedError,
            match="Oumi storage type is not supported in open source.",
        ):
            _ = reader.read(data_source)


def test_read_from_local_jsonl(reader, sample_data, sample_dataframe):
    """Test reading from local JSONL file."""
    data_source = DatasetSource(path="data/file.jsonl")

    with patch("pandas.read_json", return_value=sample_dataframe) as mock_read:
        result = reader.read(data_source)

        mock_read.assert_called_once_with("data/file.jsonl", lines=True)
        assert result == sample_data


def test_read_from_local_csv(reader, sample_data, sample_dataframe):
    """Test reading from local CSV file."""
    data_source = DatasetSource(path="data/file.csv")

    with patch("pandas.read_csv", return_value=sample_dataframe) as mock_read:
        result = reader.read(data_source)

        mock_read.assert_called_once_with("data/file.csv", sep=",")
        assert result == sample_data


def test_read_from_local_tsv(reader, sample_data, sample_dataframe):
    """Test reading from local TSV file."""
    data_source = DatasetSource(path="data/file.tsv")

    with patch("pandas.read_csv", return_value=sample_dataframe) as mock_read:
        result = reader.read(data_source)

        mock_read.assert_called_once_with("data/file.tsv", sep="\t")
        assert result == sample_data


def test_read_from_local_parquet(reader, sample_data, sample_dataframe):
    """Test reading from local Parquet file."""
    data_source = DatasetSource(path="data/file.parquet")

    with patch("pandas.read_parquet", return_value=sample_dataframe) as mock_read:
        result = reader.read(data_source)

        mock_read.assert_called_once_with("data/file.parquet")
        assert result == sample_data


def test_read_from_local_json(reader, sample_data, sample_dataframe):
    """Test reading from local JSON file."""
    data_source = DatasetSource(path="data/file.json")

    with patch("pandas.read_json", return_value=sample_dataframe) as mock_read:
        result = reader.read(data_source)

        mock_read.assert_called_once_with("data/file.json")
        assert result == sample_data


def test_read_from_local_glob_pattern(reader, sample_data):
    """Test reading from local files using glob pattern."""
    data_source = DatasetSource(path="data/*.jsonl")

    with patch.object(reader, "_read_from_glob", return_value=sample_data) as mock_glob:
        result = reader.read(data_source)

        mock_glob.assert_called_once_with("data/*.jsonl", "jsonl")
        assert result == sample_data


def test_read_with_attribute_mapping(reader, sample_data):
    """Test reading with attribute mapping."""
    data_source = DatasetSource(
        path="hf:test/dataset", attribute_map={"id": "new_id", "name": "full_name"}
    )

    with patch("oumi.core.synthesis.dataset_ingestion.load_dataset") as mock_load:
        mock_dataset = MagicMock()
        mock_dataset.__iter__ = MagicMock(return_value=iter(sample_data))
        mock_load.return_value = mock_dataset

        result = reader.read(data_source)

        expected = [
            {"new_id": 1, "full_name": "Alice"},
            {"new_id": 2, "full_name": "Bob"},
            {"new_id": 3, "full_name": "Charlie"},
        ]
        assert result == expected


def test_read_unsupported_storage_type(reader):
    """Test reading with unsupported storage type."""
    data_source = DatasetSource(path="data/file.json")
    data_source.path = "unsupported:data/file.txt"

    # Mock the storage type detection to return an invalid type
    with patch.object(DatasetPath, "_get_storage_type") as mock_get_type:
        mock_get_type.return_value = "UNSUPPORTED"

        with pytest.raises(ValueError, match="Unsupported storage type"):
            reader.read(data_source)


def test_read_unsupported_local_file_type(reader):
    """Test reading unsupported local file type."""
    data_source = DatasetSource(path="data/file.json")
    data_source.path = "data/file.txt"

    with pytest.raises(ValueError, match="Invalid path: data/file.txt"):
        reader.read(data_source)


def test_read_from_glob_multiple_files(reader, sample_data):
    """Test reading from multiple files using glob pattern."""
    data_source = DatasetSource(path="data/*.jsonl")

    # Mock Path.glob to return multiple files
    mock_files = [
        MagicMock(as_posix=MagicMock(return_value="data/file1.jsonl")),
        MagicMock(as_posix=MagicMock(return_value="data/file2.jsonl")),
    ]
    with patch("pathlib.Path.glob", return_value=mock_files):
        with (
            patch.object(
                reader,
                "_read_from_local",
                side_effect=[
                    sample_data[:2],  # First file has 2 records
                    sample_data[2:],  # Second file has 1 record
                ],
            ) as mock_read_local
        ):
            result = reader.read(data_source)
            assert mock_read_local.call_count == 2
            mock_read_local.assert_any_call("data/file1.jsonl", "jsonl")
            mock_read_local.assert_any_call("data/file2.jsonl", "jsonl")
            assert result == sample_data


def test_map_attributes_basic(reader):
    """Test basic attribute mapping."""
    samples = [
        {"old_key": "value1", "another_key": "value2"},
        {"old_key": "value3", "another_key": "value4"},
    ]
    attribute_map = {"old_key": "new_key"}

    result = reader._map_attributes(samples, attribute_map)

    expected = [
        {"new_key": "value1"},
        {"new_key": "value3"},
    ]
    assert result == expected


def test_map_attributes_multiple_mappings(reader):
    """Test attribute mapping with multiple mappings."""
    samples = [
        {"id": 1, "name": "Alice", "age": 25},
        {"id": 2, "name": "Bob", "age": 30},
    ]
    attribute_map = {"id": "user_id", "name": "full_name", "age": "years"}

    result = reader._map_attributes(samples, attribute_map)

    expected = [
        {"user_id": 1, "full_name": "Alice", "years": 25},
        {"user_id": 2, "full_name": "Bob", "years": 30},
    ]
    assert result == expected


def test_read_empty_dataset(reader):
    """Test reading empty dataset."""
    data_source = DatasetSource(path="hf:test/empty_dataset")

    with patch("oumi.core.synthesis.dataset_ingestion.load_dataset") as mock_load:
        mock_dataset = MagicMock()
        mock_dataset.__iter__ = MagicMock(return_value=iter([]))
        mock_load.return_value = mock_dataset

        result = reader.read(data_source)

        assert result == []



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builder for creating quantization instances.

This module provides a builder pattern for creating appropriate quantization
instances based on the quantization method. It serves as a simplified alternative
to the main quantize() function for when you need direct access to quantizer instances.
"""

from oumi.quantize.base import BaseQuantization


def build_quantizer(method: str) -> BaseQuantization:
    """Create appropriate quantization instance based on method.

    Args:
        method: Quantization method name (e.g., "awq_q4_0", "bnb_4bit")

    Returns:
        Instance of appropriate quantization class

    Raises:
        ValueError: If method is not supported by any quantizer
    """
    # Import here to avoid circular imports
    from oumi.quantize.awq_quantizer import AwqQuantization
    from oumi.quantize.bnb_quantizer import BitsAndBytesQuantization

    # Determine quantizer based on method prefix
    if method.startswith("awq_"):
        return AwqQuantization()
    elif method.startswith("bnb_"):
        return BitsAndBytesQuantization()
    else:
        # Try all quantizers to find one that supports this method
        for quantizer_class in [
            AwqQuantization,
            BitsAndBytesQuantization,
        ]:
            instance = quantizer_class()
            if instance.supports_method(method):
                return instance

        available_methods = get_available_methods()
        raise ValueError(
            f"Unsupported quantization method: {method}. "
            f"Available methods: {available_methods}"
        )


def get_available_methods() -> dict[str, list[str]]:
    """Returns all available methods grouped by quantization type.

    Returns:
        Dictionary mapping quantizer names to their supported methods
    """
    # Import here to avoid circular imports
    from oumi.quantize.awq_quantizer import AwqQuantization
    from oumi.quantize.bnb_quantizer import BitsAndBytesQuantization

    return {
        "AWQ": AwqQuantization.supported_methods,
        "BitsAndBytes": BitsAndBytesQuantization.supported_methods,
    }


def get_supported_formats() -> list[str]:
    """Returns all supported output formats.

    Returns:
        List of all supported output formats
    """
    # Import here to avoid circular imports
    from oumi.quantize.awq_quantizer import AwqQuantization
    from oumi.quantize.bnb_quantizer import BitsAndBytesQuantization

    formats = set()
    formats.update(AwqQuantization.supported_formats)
    formats.update(BitsAndBytesQuantization.supported_formats)

    return sorted(list(formats))


def list_all_methods() -> list[str]:
    """List all available quantization methods.

    Returns:
        Sorted list of all available method names
    """
    methods = []
    for method_list in get_available_methods().values():
        methods.extend(method_list)
    return sorted(methods)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
from typing import Any, Union

from oumi.core.configs.params.synthesis_params import (
    GeneralSynthesisParams,
    TransformationStrategy,
    TransformationType,
    TransformedAttribute,
)
from oumi.core.synthesis.attribute_formatter import AttributeFormatter
from oumi.core.types.conversation import Conversation, Message

SampleValue = Union[str, list[str], dict[str, str], Conversation]


class AttributeTransformer:
    """Transforms attributes of a dataset plan to a particular format."""

    def __init__(self, params: GeneralSynthesisParams):
        """Initializes the attribute transformer.

        Args:
            params: The general synthesis parameters containing the transformed
            attributes.
        """
        self._formatter = AttributeFormatter(params)
        self._transformed_attributes = (
            params.transformed_attributes if params.transformed_attributes else []
        )

    def transform(
        self,
        samples: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Transforms attributes of a dataset plan to a particular format.

        Args:
            samples: The samples to add transformed attributes to, using the values in
            each sample as the input to the transformation.

        Returns:
            The samples with the transformed attributes added.
        """
        for attribute in self._transformed_attributes:
            transformed_attribute_id = attribute.id
            for sample in samples:
                sample[transformed_attribute_id] = self._transform_attribute(
                    sample,
                    attribute,
                )

        return samples

    def _transform_attribute(
        self,
        sample: dict[str, Any],
        attribute: TransformedAttribute,
    ) -> SampleValue:
        """Transforms an attribute of a sample to a particular format."""
        strategy = attribute.get_strategy()
        if strategy.type == TransformationType.STRING:
            assert strategy.string_transform is not None  # Validated in __post_init__
            return self._transform_string(sample, strategy.string_transform)
        elif strategy.type == TransformationType.LIST:
            return self._transform_list(sample, strategy)
        elif strategy.type == TransformationType.DICT:
            return self._transform_dict(sample, strategy)
        elif strategy.type == TransformationType.CHAT:
            return self._transform_chat(sample, strategy, attribute.id)
        else:
            raise ValueError(f"Unsupported transformation strategy: {strategy.type}")

    def _transform_string(
        self,
        sample: dict[str, SampleValue],
        transform: str,
    ) -> str:
        """Transforms a string attribute of a sample to a particular format."""
        string_sample = {k: v for k, v in sample.items() if isinstance(v, str)}
        formatted_string = self._formatter.format(
            string_sample,
            transform,
            missing_values_allowed=False,
        )
        return formatted_string

    def _transform_list(
        self,
        sample: dict[str, SampleValue],
        transform: TransformationStrategy,
    ) -> list[str]:
        """Transforms a list attribute of a sample to a particular format."""
        assert transform.list_transform is not None
        return [self._transform_string(sample, e) for e in transform.list_transform]

    def _transform_dict(
        self,
        sample: dict[str, SampleValue],
        transform: TransformationStrategy,
    ) -> dict[str, str]:
        """Transforms a dict attribute of a sample to a particular format."""
        assert transform.dict_transform is not None  # Validated in __post_init__
        return {
            k: self._transform_string(sample, v)
            for k, v in transform.dict_transform.items()
        }

    def _transform_chat(
        self,
        sample: dict[str, SampleValue],
        transform: TransformationStrategy,
        attribute_id: str,
    ) -> dict[str, Any]:
        """Transforms a chat attribute of a sample to a particular format."""
        assert transform.chat_transform is not None  # Validated in __post_init__
        messages = []
        for message in transform.chat_transform.messages:
            content = message.content
            if not isinstance(content, str):
                raise ValueError(
                    "ChatTransform.transforms.messages.content must be a string."
                )

            formatted_content = self._transform_string(sample, content)
            messages.append(Message(role=message.role, content=formatted_content))

        transformed_metadata = {}
        if transform.chat_transform.metadata:
            # Create a TransformationStrategy for the metadata dict transformation
            metadata_transform = TransformationStrategy(
                type=TransformationType.DICT,
                dict_transform=transform.chat_transform.metadata,
            )
            transformed_metadata = self._transform_dict(sample, metadata_transform)

        new_conv_id = transform.chat_transform.conversation_id
        if not transform.chat_transform.conversation_id:
            new_conv_id = f"{attribute_id}-{uuid.uuid4()}"

        return Conversation(
            messages=messages,
            conversation_id=new_conv_id,
            metadata=transformed_metadata,
        ).to_dict()



# Copyright 2020-2025 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Any, Optional, Union

import numpy as np
import torch
from transformers.data.data_collator import DataCollatorForLanguageModeling


class DataCollatorForCompletionOnlyLM(DataCollatorForLanguageModeling):
    """Data collator used for completion tasks.

    Copied from `trl`'s `DataCollatorForCompletionOnlyLM` class.
    """

    def __init__(
        self,
        response_template: Union[str, list[int]],
        instruction_template: Optional[Union[str, list[int]]] = None,
        *args,
        mlm: bool = False,
        ignore_index: int = -100,
        padding_free: bool = False,
        **kwargs,
    ):
        """Initializes the DataCollatorForCompletionOnlyLM."""
        super().__init__(*args, mlm=mlm, **kwargs)

        self.instruction_template = instruction_template
        if isinstance(instruction_template, str):
            # The user provides a string, must tokenize
            self.instruction_token_ids = self.tokenizer.encode(
                self.instruction_template,  # type: ignore
                add_special_tokens=False,
            )
        else:
            # The user already provides the token ids
            self.instruction_token_ids = instruction_template

        self.response_template = response_template
        if isinstance(response_template, str):
            # The user provides a string, must tokenize
            self.response_token_ids = self.tokenizer.encode(
                self.response_template, add_special_tokens=False
            )
        else:
            # The user already provides the token ids
            self.response_token_ids = response_template

        if (
            not self.mlm
            and self.instruction_template
            and self.tokenizer.pad_token_id == self.tokenizer.eos_token_id
        ):
            warnings.warn(
                "The pad_token_id and eos_token_id values of this tokenizer "
                "are identical. "
                "If you are planning for multi-turn training, "
                "it can result in the model continuously generating questions and "
                "answers without eos token. "
                "To avoid this, set the pad_token_id to a different value.",
                UserWarning,
            )

        self.ignore_index = ignore_index
        self.padding_free = padding_free

    def torch_call(
        self, examples: list[Union[list[int], Any, dict[str, Any]]]
    ) -> dict[str, Any]:
        """Collates a list of examples into a batch."""
        batch = super().torch_call(examples)

        if self.instruction_template is None:
            for i in range(len(examples)):
                response_token_ids_start_idx = None

                for idx in np.where(batch["labels"][i] == self.response_token_ids[0])[
                    0
                ]:
                    # `response_token_ids` is `'### Response:\n'`,
                    # here we are just making sure that the token IDs match
                    if (
                        self.response_token_ids
                        == batch["labels"][i][
                            idx : idx + len(self.response_token_ids)
                        ].tolist()
                    ):
                        response_token_ids_start_idx = idx

                if response_token_ids_start_idx is None:
                    warnings.warn(
                        f"Could not find response key `{self.response_template}` "
                        "in the following instance: "
                        f"{self.tokenizer.decode(batch['input_ids'][i])}. "
                        "This instance will be ignored in loss "
                        "calculation. Note, if this happens often, consider "
                        "increasing the `max_length`.",
                        UserWarning,
                    )
                    batch["labels"][i, :] = self.ignore_index
                else:
                    response_token_ids_end_idx = response_token_ids_start_idx + len(
                        self.response_token_ids
                    )

                    # Make pytorch loss function ignore all tokens up through the end
                    # of the response key
                    batch["labels"][i, :response_token_ids_end_idx] = self.ignore_index

        else:
            for i in range(len(examples)):
                response_token_ids_idxs = []
                human_token_ids_idxs = []

                for assistant_idx in np.where(
                    batch["labels"][i] == self.response_token_ids[0]
                )[0]:
                    # find the indexes of the start of a response.
                    if (
                        self.response_token_ids
                        == batch["labels"][i][
                            assistant_idx : assistant_idx + len(self.response_token_ids)
                        ].tolist()
                    ):
                        response_token_ids_idxs.append(
                            assistant_idx + len(self.response_token_ids)
                        )

                if len(response_token_ids_idxs) == 0:
                    warnings.warn(
                        f"Could not find response key `{self.response_template}` "
                        "in the following instance: "
                        f"{self.tokenizer.decode(batch['input_ids'][i])}. "
                        "This instance will be ignored in loss "
                        "calculation. Note, if this happens often, consider "
                        "increasing the `max_length`.",
                        UserWarning,
                    )
                    batch["labels"][i, :] = self.ignore_index

                human_token_ids = self.instruction_token_ids
                for human_idx in np.where(batch["labels"][i] == human_token_ids[0])[0]:  # type: ignore
                    # find the indexes of the start of a human answer.
                    if (
                        human_token_ids
                        == batch["labels"][i][
                            human_idx : human_idx + len(human_token_ids)
                        ].tolist()
                    ):
                        human_token_ids_idxs.append(human_idx)

                if len(human_token_ids_idxs) == 0:
                    warnings.warn(
                        f"Could not find instruction key `{self.instruction_template}` "
                        "in the following instance: "
                        f"{self.tokenizer.decode(batch['input_ids'][i])}."
                        " This instance will be ignored in loss "
                        "calculation. Note, if this happens often, "
                        "consider increasing the `max_length`.",
                        UserWarning,
                    )
                    batch["labels"][i, :] = self.ignore_index

                if (
                    len(human_token_ids_idxs) > 0
                    and len(response_token_ids_idxs) > 0
                    and human_token_ids_idxs[0] > response_token_ids_idxs[0]
                ):
                    human_token_ids_idxs = [0] + human_token_ids_idxs

                for idx, (start, end) in enumerate(
                    zip(human_token_ids_idxs, response_token_ids_idxs)
                ):
                    # Make pytorch loss function ignore all non response tokens
                    if idx != 0:
                        batch["labels"][i, start:end] = self.ignore_index
                    else:
                        batch["labels"][i, :end] = self.ignore_index

                if len(response_token_ids_idxs) < len(human_token_ids_idxs):
                    batch["labels"][i, human_token_ids_idxs[-1] :] = self.ignore_index

        if self.padding_free:
            # remove padding, `attention_mask` and add `position_ids`
            attn_mask = batch.pop("attention_mask")
            batch["input_ids"] = batch["input_ids"][attn_mask.bool()].unsqueeze(0)
            batch["position_ids"] = (
                attn_mask.cumsum(1)[attn_mask.bool()].unsqueeze(0) - 1
            )
            batch["labels"] = batch["labels"][attn_mask.bool()].unsqueeze(0)
            batch["labels"][batch["position_ids"] == 0] = self.ignore_index

            # Calculate cumulative sequence lengths for queries and keys to prevent
            #  graph breaks during further computations.
            flattened_position_ids = batch["position_ids"].flatten()
            indices_q = torch.arange(
                flattened_position_ids.size(0),
                device=flattened_position_ids.device,
                dtype=torch.int32,
            )
            batch["cu_seq_lens_q"] = torch.cat(
                (
                    indices_q[flattened_position_ids == 0],
                    torch.tensor(
                        flattened_position_ids.size(),
                        device=flattened_position_ids.device,
                        dtype=torch.int32,
                    ),
                )
            ).unsqueeze(0)
            batch["cu_seq_lens_k"] = batch["cu_seq_lens_q"]

            # Determine maximum sequence lengths to prevent graph breaks during
            #  further computations.
            batch["max_length_k"] = torch.tensor(
                [flattened_position_ids.max().item() + 1]
            )
            batch["max_length_q"] = batch["max_length_k"]

        return batch



import tempfile
from pathlib import Path
from typing import Optional
from unittest import mock
from unittest.mock import patch

import jsonlines
import pytest

from oumi.core.configs import GenerationParams, InferenceConfig, ModelParams
from oumi.core.inference import BaseInferenceEngine
from oumi.core.types.conversation import Conversation, Message, Role


class MockInferenceEngine(BaseInferenceEngine):
    """Mock inference engine for testing scratch file functionality."""

    def get_supported_params(self) -> set[str]:
        return {"max_new_tokens", "temperature"}

    def _infer_online(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        # Mock implementation that appends an assistant response
        results = []
        for i, conv in enumerate(input):
            new_conv = conv.model_copy(deep=True)
            new_conv.messages.append(
                Message(
                    role=Role.ASSISTANT,
                    content=f"Mock response {i}",
                )
            )
            results.append(new_conv)
            self._save_conversation_to_scratch(
                new_conv,
                inference_config.output_path if inference_config else None,
            )
        return results

    def _infer_from_file(
        self,
        input_filepath: str,
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        return self.infer(inference_config=inference_config)


@pytest.fixture
def mock_engine():
    model_params = ModelParams(model_name="test-model")
    return MockInferenceEngine(model_params=model_params)


def create_test_conversation(idx: int) -> Conversation:
    """Creates a test conversation with a unique ID."""
    return Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=f"Test message {idx}",
            )
        ],
        conversation_id=f"test-{idx}",
    )


def test_scratch_file_creation_and_cleanup(mock_engine):
    """Test that scratch files are created and cleaned up properly."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        conversations = [create_test_conversation(1)]

        # Run inference with a patched _cleanup_scratch_file to prevent cleanup
        with patch.object(mock_engine, "_cleanup_scratch_file") as mock_cleanup:
            results = mock_engine.infer(
                input=conversations,
                inference_config=inference_config,
            )

            # Get the actual scratch path used by the engine
            scratch_path = Path(mock_engine._get_scratch_filepath(output_path))

            # Verify scratch file was created and exists
            assert scratch_path.exists(), "Scratch file should exist after inference"

            # Verify results
            assert len(results) == 1
            assert results[0].conversation_id == "test-1"
            assert len(results[0].messages) == 2  # Original + assistant response

        # Verify that cleanup was called
        mock_cleanup.assert_called_once_with(output_path)

        # Manually call cleanup since we prevented the automatic cleanup from the patch
        mock_engine._cleanup_scratch_file(output_path)


def test_infer_no_resume_from_scratch_on_success(mock_engine):
    """Test that inference processes all conversations each time since scratch is
    cleaned up after successful inference."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        # Create two conversations
        conversations = [
            create_test_conversation(1),
            create_test_conversation(2),
        ]

        with patch.object(
            mock_engine,
            "_infer_online",
            wraps=mock_engine._infer_online,
        ) as mock_infer:
            # Process only first conversation
            with patch.object(
                mock_engine,
                "_save_conversation_to_scratch",
                wraps=mock_engine._save_conversation_to_scratch,
            ) as mock_save:
                first_results = mock_engine.infer(
                    input=[conversations[0].model_copy(deep=True)],
                    inference_config=inference_config,
                )
                # Verify scratch file was created and saved to
                assert mock_save.called

            # Get scratch path and verify it was cleaned up after first inference
            scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
            assert not scratch_path.exists()

            # Process all conversations
            with patch.object(
                mock_engine,
                "_save_conversation_to_scratch",
                wraps=mock_engine._save_conversation_to_scratch,
            ) as mock_save:
                all_results = mock_engine.infer(
                    input=conversations,  # Pass both conversations
                    inference_config=inference_config,
                )
                # Verify scratch file was created and saved to
                assert mock_save.called

            # Verify results
            assert len(first_results) == 1
            assert len(all_results) == 2
            assert all_results[0].conversation_id == "test-1"
            assert all_results[1].conversation_id == "test-2"

            # Each conversation should have the original message + assistant response
            assert len(all_results[0].messages) == 2
            assert len(all_results[1].messages) == 2

            # Verify infer_online was called with all inputs each time (no resuming)
            mock_infer.assert_has_calls(
                [
                    # First call with just the first conversation
                    mock.call(
                        [conversations[0].model_copy(deep=True)], inference_config
                    ),
                    # Second call with both conversations
                    mock.call(conversations, inference_config),
                ]
            )

            # Get scratch path and verify it was cleaned up after final inference
            scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
            assert not scratch_path.exists()


def test_infer_resume_from_scratch_on_failure(mock_engine):
    """Test that inference resumes from scratch if it fails."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        # Create two conversations
        conversations = [
            create_test_conversation(1),
            create_test_conversation(2),
        ]

        # Run inference which fails on the second conversation
        def mock_infer_online(input_convs, config):
            # Process conversations one at a time
            results = []
            for i, conv in enumerate(input_convs):
                if i == 1:  # Fail on second conversation
                    raise RuntimeError("Failed processing second conversation")
                # Process first conversation normally
                new_conv = conv.model_copy(deep=True)
                new_conv.messages.append(
                    Message(
                        role=Role.ASSISTANT,
                        content=f"Mock response {conv.conversation_id}",
                    )
                )
                results.append(new_conv)
                # Save first conversation to scratch
                mock_engine._save_conversation_to_scratch(
                    new_conv,
                    config.output_path if config else None,
                )
            return results

        with patch.object(
            mock_engine,
            "_infer_online",
            side_effect=mock_infer_online,
        ) as mock_infer:
            # Should fail on second conversation
            with pytest.raises(RuntimeError):
                mock_engine.infer(
                    input=conversations,
                    inference_config=inference_config,
                )

            # Get scratch path and verify it exists and contains first conversation
            scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
            assert scratch_path.exists()

            # Verify infer_online was called with both conversations
            mock_infer.assert_called_once()
            assert len(mock_infer.call_args[0][0]) == 2

            # Verify that the scratch file contains the first conversation
            with open(scratch_path) as f:
                lines = f.readlines()
                assert len(lines) == 1
                first_conv = Conversation.from_json(lines[0])
                assert (
                    first_conv.messages[0].content
                    == conversations[0].messages[0].content
                )
                assert first_conv.messages[-1].role == Role.ASSISTANT
                assert first_conv.messages[-1].content == "Mock response test-1"

        # Run inference again, this time with no errors
        # It should resume from scratch, only processing the second conversation
        with patch.object(
            mock_engine,
            "_infer_online",
            side_effect=mock_infer_online,
        ) as mock_infer:
            results = mock_engine.infer(
                input=conversations,
                inference_config=inference_config,
            )

            # Verify that results contain both conversations
            assert len(results) == 2
            assert results[0].conversation_id == "test-1"
            assert results[1].conversation_id == "test-2"
            assert len(results[0].messages) == 2
            assert len(results[1].messages) == 2

            # Get the actual scratch path and verify that it was cleaned up
            scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
            assert not scratch_path.exists()

            # Verify that infer_online was called with only the second conversation
            mock_infer.assert_called_once()
            assert len(mock_infer.call_args[0][0]) == 1
            assert mock_infer.call_args[0][0][0].conversation_id == "test-2"

        # Verify that output file exists and contains expected conversations
        output_file_path = Path(output_path)
        assert output_file_path.exists(), (
            "Output file should exist after successful inference"
        )

        saved_conversations = []
        with jsonlines.open(output_path) as reader:
            for obj in reader:
                saved_conversations.append(Conversation.from_dict(obj))

        assert len(saved_conversations) == 2, (
            "Output file should contain all processed conversations"
        )

        # Verify that the saved conversations match the results
        for i, (result_conv, saved_conv) in enumerate(
            zip(results, saved_conversations)
        ):
            assert result_conv.conversation_id == saved_conv.conversation_id
            assert len(saved_conv.messages) == 2, (
                f"Conversation {i} should have original + assistant message"
            )

            # Verify original user message
            assert saved_conv.messages[0].role == Role.USER
            assert saved_conv.messages[0].content == f"Test message {i + 1}"

            # Verify assistant response was added
            assert saved_conv.messages[1].role == Role.ASSISTANT
            assert saved_conv.messages[1].content == f"Mock response test-{i + 1}"


def test_scratch_file_handling_with_errors(mock_engine):
    """Test that scratch files are handled properly even when errors occur."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        conversations = [create_test_conversation(1)]

        # Simulate an error during inference
        with patch.object(
            mock_engine, "_infer_online", side_effect=RuntimeError("Test error")
        ):
            with pytest.raises(RuntimeError):
                mock_engine.infer(
                    input=conversations,
                    inference_config=inference_config,
                )

        # Get the actual scratch path and verify it was cleaned up despite the error
        scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
        assert not scratch_path.exists()


def test_empty_scratch_file(mock_engine):
    """Ensure that inference works even if the scratch file is empty."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        # Create scratch directory and empty file using the engine's actual path
        actual_scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
        actual_scratch_path.parent.mkdir(parents=True, exist_ok=True)
        actual_scratch_path.touch()

        conversations = [create_test_conversation(1)]

        # Run inference
        results = mock_engine.infer(
            input=conversations,
            inference_config=inference_config,
        )

        # Verify results
        assert len(results) == 1
        assert results[0].conversation_id == "test-1"
        assert len(results[0].messages) == 2  # Original + assistant response


def test_full_scratch_file(mock_engine):
    """Validate that inference doesn't run if scratch file has all conversations."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        conversations = [create_test_conversation(1)]

        # Create scratch file with all conversations using engine's actual path
        mock_engine._dataset_hash = (
            "8a03cc24121fd45c48ee2950b404ff17f66caa7a04284cd9b8b7aab9cf63996e"
        )
        actual_scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
        actual_scratch_path.parent.mkdir(parents=True, exist_ok=True)
        actual_scratch_path.touch()
        with jsonlines.open(actual_scratch_path, "w") as writer:
            writer.write(conversations[0].to_dict())

        # Run inference
        results = mock_engine.infer(
            input=conversations,
            inference_config=inference_config,
        )

        # Verify results
        assert len(results) == 1
        assert results[0].conversation_id == "test-1"
        assert len(results[0].messages) == 1  # Original, no assistant response in file


def test_final_conversations_saved_to_output_file(mock_engine):
    """Test that final conversations are saved to output file after inference."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = str(Path(temp_dir) / "output.jsonl")
        inference_config = InferenceConfig(
            output_path=output_path,
            generation=GenerationParams(max_new_tokens=10),
        )

        # Create test conversations
        conversations = [
            create_test_conversation(1),
            create_test_conversation(2),
        ]

        results = mock_engine.infer(
            input=conversations,
            inference_config=inference_config,
        )

        output_file_path = Path(output_path)
        assert output_file_path.exists(), (
            "Output file should exist after successful inference"
        )

        saved_conversations = []
        with jsonlines.open(output_path) as reader:
            for obj in reader:
                saved_conversations.append(Conversation.from_dict(obj))
        assert len(saved_conversations) == 2, (
            "Output file should contain all processed conversations"
        )

        # Verify that the saved conversations match the results
        for i, (result_conv, saved_conv) in enumerate(
            zip(results, saved_conversations)
        ):
            assert result_conv.conversation_id == saved_conv.conversation_id
            assert len(saved_conv.messages) == 2, (
                f"Conversation {i} should have original + assistant message"
            )

            # Verify original user message
            assert saved_conv.messages[0].role == Role.USER
            assert saved_conv.messages[0].content == f"Test message {i + 1}"

            # Verify assistant response was added
            assert saved_conv.messages[1].role == Role.ASSISTANT
            assert saved_conv.messages[1].content == f"Mock response {i}"

        # Get the actual scratch path and verify that it was cleaned up
        scratch_path = Path(mock_engine._get_scratch_filepath(output_path))
        assert not scratch_path.exists(), (
            "Scratch file should be cleaned up after successful inference"
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import traceback

import typer

from oumi.cli.cache import card as cache_card
from oumi.cli.cache import get as cache_get
from oumi.cli.cache import ls as cache_ls
from oumi.cli.cache import rm as cache_rm
from oumi.cli.cli_utils import (
    CONSOLE,
    CONTEXT_ALLOW_EXTRA_ARGS,
    create_github_issue_url,
)
from oumi.cli.distributed_run import accelerate, torchrun
from oumi.cli.env import env
from oumi.cli.evaluate import evaluate
from oumi.cli.fetch import fetch
from oumi.cli.infer import infer
from oumi.cli.judge import judge_conversations_file, judge_dataset_file
from oumi.cli.launch import cancel, down, logs, status, stop, up, which
from oumi.cli.launch import run as launcher_run
from oumi.cli.quantize import quantize
from oumi.cli.synth import synth
from oumi.cli.train import train
from oumi.cli.tune import tune
from oumi.utils.logging import should_use_rich_logging

_ASCII_LOGO = r"""
   ____  _    _ __  __ _____
  / __ \| |  | |  \/  |_   _|
 | |  | | |  | | \  / | | |
 | |  | | |  | | |\/| | | |
 | |__| | |__| | |  | |_| |_
  \____/ \____/|_|  |_|_____|
"""


def experimental_features_enabled():
    """Check if experimental features are enabled."""
    is_enabled = os.environ.get("OUMI_ENABLE_EXPERIMENTAL_FEATURES", "False")
    return is_enabled.lower() in ("1", "true", "yes", "on")


def _oumi_welcome(ctx: typer.Context):
    if ctx.invoked_subcommand == "distributed":
        return
    # Skip logo for rank>0 for multi-GPU jobs to reduce noise in logs.
    if int(os.environ.get("RANK", 0)) > 0:
        return
    CONSOLE.print(_ASCII_LOGO, style="green", highlight=False)


def get_app() -> typer.Typer:
    """Create the Typer CLI app."""
    app = typer.Typer(pretty_exceptions_enable=False)
    app.callback(context_settings={"help_option_names": ["-h", "--help"]})(
        _oumi_welcome
    )
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Evaluate a model.",
    )(evaluate)
    app.command()(env)
    app.command(  # Alias for evaluate
        name="eval",
        hidden=True,
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Evaluate a model.",
    )(evaluate)
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Run inference on a model.",
    )(infer)
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Synthesize a dataset.",
    )(synth)
    app.command(  # Alias for synth
        name="synthesize",
        hidden=True,
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="🚧 [Experimental] Synthesize a dataset.",
    )(synth)
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Train a model.",
    )(train)
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Tune the parameters for a model.",
    )(tune)
    app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS,
        help="Quantize a model.",
    )(quantize)
    judge_app = typer.Typer(pretty_exceptions_enable=False)
    judge_app.command(name="dataset", context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(
        judge_dataset_file
    )
    judge_app.command(name="conversations", context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(
        judge_conversations_file
    )
    app.add_typer(judge_app, name="judge", help="Judge datasets or conversations.")

    launch_app = typer.Typer(pretty_exceptions_enable=False)
    launch_app.command(help="Cancels a job.")(cancel)
    launch_app.command(help="Turns down a cluster.")(down)
    launch_app.command(
        name="run", context_settings=CONTEXT_ALLOW_EXTRA_ARGS, help="Runs a job."
    )(launcher_run)
    launch_app.command(help="Prints the status of jobs launched from Oumi.")(status)
    launch_app.command(help="Stops a cluster.")(stop)
    launch_app.command(
        context_settings=CONTEXT_ALLOW_EXTRA_ARGS, help="Launches a job."
    )(up)
    launch_app.command(help="Prints the available clouds.")(which)
    launch_app.command(help="Gets the logs of a job.")(logs)
    app.add_typer(launch_app, name="launch", help="Launch jobs remotely.")

    distributed_app = typer.Typer(pretty_exceptions_enable=False)
    distributed_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(accelerate)
    distributed_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(torchrun)
    app.add_typer(
        distributed_app,
        name="distributed",
        help=(
            "A wrapper for torchrun/accelerate "
            "with reasonable default values for distributed training."
        ),
    )

    app.command(
        help="Fetch configuration files from the oumi GitHub repository.",
    )(fetch)

    cache_app = typer.Typer(pretty_exceptions_enable=False)
    cache_app.command(name="ls", help="List locally cached items.")(cache_ls)
    cache_app.command(name="get", help="Download a repository from Hugging Face.")(
        cache_get
    )
    cache_app.command(name="card", help="Show information for a repository.")(
        cache_card
    )
    cache_app.command(name="rm", help="Remove a repository from the local cache.")(
        cache_rm
    )
    app.add_typer(cache_app, name="cache", help="Manage local Hugging Face cache.")

    return app


def run():
    """The entrypoint for the CLI."""
    app = get_app()
    try:
        return app()
    except Exception as e:
        tb_str = traceback.format_exc()
        CONSOLE.print(tb_str)
        issue_url = create_github_issue_url(e, tb_str)
        CONSOLE.print(
            "\n[red]If you believe this is a bug, please file an issue:[/red]"
        )
        if should_use_rich_logging():
            CONSOLE.print(
                f"📝 [yellow]Templated issue:[/yellow] "
                f"[link={issue_url}]Click here to report[/link]"
            )
        else:
            CONSOLE.print(
                "https://github.com/oumi-ai/oumi/issues/new?template=bug-report.yaml"
            )

        sys.exit(1)


if "sphinx" in sys.modules:
    # Create the CLI app when building the docs to auto-generate the CLI reference.
    app = get_app()



from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.core.registry import REGISTRY, RegistryType
from oumi.launcher.clients.polaris_client import PolarisClient
from oumi.launcher.clouds.polaris_cloud import PolarisCloud
from oumi.launcher.clusters.polaris_cluster import PolarisCluster


#
# Fixtures
#
@pytest.fixture
def mock_polaris_client():
    with patch("oumi.launcher.clouds.polaris_cloud.PolarisClient") as client:
        client.SupportedQueues = PolarisClient.SupportedQueues
        yield client


@pytest.fixture
def mock_polaris_cluster():
    with patch("oumi.launcher.clouds.polaris_cloud.PolarisCluster") as cluster:
        yield cluster


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_polaris_cloud_up_cluster_debug(mock_polaris_client, mock_polaris_cluster):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="debug.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, "debug.user")
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_up_cluster_demand(mock_polaris_client, mock_polaris_cluster):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="demand.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, "demand.user")
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_up_cluster_debug_scaling(
    mock_polaris_client, mock_polaris_cluster
):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="debug-scaling.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, "debug-scaling.user")
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_up_cluster_preemptable(
    mock_polaris_client, mock_polaris_cluster
):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="preemptable.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, "preemptable.user")
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_up_cluster_prod(mock_polaris_client, mock_polaris_cluster):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="prod.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, "prod.user")
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_up_cluster_fails_mismatched_user(
    mock_polaris_client, mock_polaris_cluster
):
    cloud = PolarisCloud()
    with pytest.raises(ValueError):
        _ = cloud.up_cluster(_get_default_job("polaris"), "debug.user1")


def test_polaris_cloud_up_cluster_default_queue(
    mock_polaris_client, mock_polaris_cluster
):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=PolarisCluster)
    mock_polaris_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="prod.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("polaris")
    job_status = cloud.up_cluster(job, None)
    mock_polaris_client.assert_called_once_with("user")
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_polaris_cloud_init_with_users(mock_polaris_client):
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client, mock_client]
    mock_polaris_client.get_active_users.return_value = ["user1", "user2"]
    cloud = PolarisCloud()
    cluster_names = [cluster.name() for cluster in cloud.list_clusters()]
    cluster_names.sort()
    assert cluster_names == [
        "debug-scaling.user1",
        "debug-scaling.user2",
        "debug.user1",
        "debug.user2",
        "demand.user1",
        "demand.user2",
        "preemptable.user1",
        "preemptable.user2",
        "prod.user1",
        "prod.user2",
    ]
    mock_polaris_client.assert_has_calls([call("user1"), call("user2")])


def test_polaris_cloud_initialize_cluster(mock_polaris_client):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client]
    clusters = cloud.initialize_clusters("me")
    clusters2 = cloud.initialize_clusters("me")
    mock_polaris_client.assert_called_once_with("me")
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == [
        "debug-scaling.me",
        "debug.me",
        "demand.me",
        "preemptable.me",
        "prod.me",
    ]
    # Verify that the second initialization returns the same clusters.
    assert clusters == clusters2


def test_polaris_cloud_list_clusters(mock_polaris_client):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client, mock_client]
    # Check that there are no initial clusters.
    assert [] == cloud.list_clusters()
    cloud.initialize_clusters("me")
    clusters = cloud.list_clusters()
    expected_clusters = [
        "debug-scaling.me",
        "debug.me",
        "demand.me",
        "preemptable.me",
        "prod.me",
    ]
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == expected_clusters


def test_polaris_cloud_get_cluster_empty(mock_polaris_client):
    cloud = PolarisCloud()
    # Check that there are no initial clusters.
    assert cloud.get_cluster("debug.user") is None


def test_polaris_cloud_get_cluster_success(mock_polaris_client):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client, mock_client]
    cloud.initialize_clusters("me")
    expected_clusters = [
        "debug-scaling.me",
        "debug.me",
        "demand.me",
        "preemptable.me",
        "prod.me",
    ]
    for name in expected_clusters:
        cluster = cloud.get_cluster(name)
        assert cluster is not None
        assert cluster.name() == name


def test_polaris_cloud_get_cluster_fails(mock_polaris_client):
    cloud = PolarisCloud()
    mock_client = Mock(spec=PolarisClient)
    mock_polaris_client.side_effect = [mock_client, mock_client]
    cloud.initialize_clusters("me")
    assert cloud.get_cluster("nonexistent") is None


def test_polaris_cloud_builder_registered():
    assert REGISTRY.contains("polaris", RegistryType.CLOUD)



import copy

from oumi.utils.logging import logger

from .clip_convnext_encoder import CLIPConvNextTower
from .clip_encoder import ClipVisionTower
from .dino_encoder import DinoVisionTower
from .siglip_encoder import SiglipVisionTower


def build_vision_tower(vision_tower_cfg, **kwargs):
    vision_tower = getattr(
        vision_tower_cfg,
        "mm_vision_tower",
        getattr(vision_tower_cfg, "vision_tower", None),
    )

    if vision_tower is None or not isinstance(vision_tower, str):
        raise ValueError(
            f"Vision Tower is not specified in the config: {vision_tower_cfg}"
        )

    # CLIP-based Vision Towers
    if "openai/clip" in vision_tower.lower():
        logger.info(f"Loading **OpenAI CLIP** Vision Tower: {vision_tower}")
        return ClipVisionTower(vision_tower, args=vision_tower_cfg, **kwargs)
    if "siglip" in vision_tower.lower():
        logger.info(f"Loading **SigLIP CLIP** Vision Tower: {vision_tower}")
        return SiglipVisionTower(vision_tower, args=vision_tower_cfg, **kwargs)
    if "clip-convnext" in vision_tower.lower():
        logger.info(f"Loading **ConvNeXt CLIP** Vision Tower: {vision_tower}")
        return CLIPConvNextTower(vision_tower, args=vision_tower_cfg, **kwargs)

    # SSL-based Vision Towers
    if "dinov2" in vision_tower.lower():
        logger.info(f"Loading **DINO Vision Tower: {vision_tower}")
        return DinoVisionTower(vision_tower, args=vision_tower_cfg, **kwargs)

    # Supervised Vision Towers

    # Other Vision Towers

    raise ValueError(f"Unknown vision tower: {vision_tower}")


def build_vision_tower_aux_list(vision_tower_cfg, **kwargs):
    vision_tower_aux_name_list = getattr(
        vision_tower_cfg,
        "mm_vision_tower_aux_list",
        getattr(vision_tower_cfg, "vision_tower_aux_list", None),
    )
    vision_tower_aux_token_len_list = getattr(
        vision_tower_cfg,
        "mm_vision_tower_aux_token_len_list",
        getattr(vision_tower_cfg, "vision_tower_aux_token_len_list", None),
    )
    vision_tower_aux_list = []
    for vision_tower_aux_name, vision_tower_aux_token_len in zip(
        vision_tower_aux_name_list, vision_tower_aux_token_len_list
    ):
        config = copy.deepcopy(vision_tower_cfg)
        vision_tower_aux_name += f"-interp{vision_tower_aux_token_len}"

        # CLIP-based Vision Towers
        if "openai/clip" in vision_tower_aux_name.lower():
            logger.info(
                f"Loading **OpenAI CLIP** Vision Tower: {vision_tower_aux_name}"
            )
            vision_tower_aux_list.append(
                ClipVisionTower(vision_tower_aux_name, args=config, **kwargs)
            )
        elif "siglip" in vision_tower_aux_name.lower():
            logger.info(
                f"Loading **SigLIP CLIP** Vision Tower: {vision_tower_aux_name}"
            )
            vision_tower_aux_list.append(
                SiglipVisionTower(vision_tower_aux_name, args=config, **kwargs)
            )
        elif "clip-convnext" in vision_tower_aux_name.lower():
            logger.info(
                f"Loading **ConvNeXt CLIP** Vision Tower: {vision_tower_aux_name}"
            )
            vision_tower_aux_list.append(
                CLIPConvNextTower(vision_tower_aux_name, args=config, **kwargs)
            )
        # SSL-based Vision Towers
        elif "dinov2" in vision_tower_aux_name.lower():
            logger.info(f"Loading **DINO Vision Tower: {vision_tower_aux_name}")
            vision_tower_aux_list.append(
                DinoVisionTower(vision_tower_aux_name, args=config, **kwargs)
            )
        # Supervised Vision Towers
        # Other Vision Towers
        else:
            raise ValueError(f"Unknown vision tower: {vision_tower_aux_name}")
    return vision_tower_aux_list



# Copyright 2025 - Jiayi Pan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Derived from https://github.com/Jiayi-Pan/TinyZero/blob/main/examples/data_preprocess/countdown.py.

This file was slightly modified to inherit from the `BaseExperimentalGrpoDataset` class.
"""

import pandas as pd
from typing_extensions import override

from oumi.core.datasets.base_grpo_dataset import BaseExperimentalGrpoDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation

_PROMPT_TEMPLATE = """A conversation between User and Assistant. The user asks a question, and the Assistant solves it. The assistant first thinks about the reasoning process in the mind and then provides the user with the answer.
User: Using the numbers {nums}, create an equation that equals {target}. You can use basic arithmetic operations (+, -, *, /) and each number can only be used once. Show your work in <think> </think> tags. And return the final answer in <answer> </answer> tags, for example <answer> (1 + 2) / 3 </answer>.
Assistant: Let me solve this step by step.
<think>"""  # noqa: E501


@register_dataset("d1shs0ap/countdown")
class CountdownGrpoDataset(BaseExperimentalGrpoDataset):
    """Dataset class for the `d1shs0ap/countdown` dataset.

    A sample from the dataset:
    {"target": 87, "nums": [79, 8]}
    """

    default_dataset = "d1shs0ap/countdown"

    @override
    def transform(self, sample: pd.Series) -> dict:
        """Validate and transform the sample into Python `dict`."""
        target = int(sample["target"])
        # Convert nums from np array to list
        nums = [int(num) for num in sample["nums"]]
        prompt = _PROMPT_TEMPLATE.format(nums=nums, target=target)
        solution = {
            "target": target,
            "numbers": nums,
        }
        return {
            "data_source": "countdown",
            "prompt": [
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            "ability": "math",
            "reward_model": {"style": "rule", "ground_truth": solution},
            "extra_info": {
                "split": self.split if self.split else "",
            },
        }

    @override
    def transform_conversation(self, sample: pd.Series) -> Conversation:
        """Validate and transform the sample into Python `dict`."""
        sample_dict = self.transform(sample)
        prompt_message = sample_dict["prompt"][0]
        del sample_dict["prompt"]
        conversation_dict = {"messages": [prompt_message], "metadata": sample_dict}
        return Conversation.from_dict(conversation_dict)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override

from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)

_GEOMETRY3K_INSTRUCTION = (
    r"You FIRST think about the reasoning process as an internal monologue and then "
    r"provide the final answer. "
    r"The reasoning process MUST BE enclosed within <think> </think> tags. "
    r"The final answer MUST BE put in \boxed{}."
)


@register_dataset("hiyouga/geometry3k")
class Geometry3kDataset(VisionLanguageSftDataset):
    """Dataset class for the `hiyouga/geometry3k` dataset."""

    default_dataset = "hiyouga/geometry3k"

    def __init__(
        self,
        *,
        add_system_instruction: bool = False,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the Geometry3kDataset class."""
        self._add_system_instruction = add_system_instruction
        super().__init__(**kwargs)

    def _process_text_value(self, s: str) -> str:
        # The data contains occasional `\n` at the beginning or end
        # of text values. Let's strip them.
        return s.strip() if s else ""

    @override
    def transform_conversation(self, example: dict) -> Conversation:
        """Transform a single conversation example into a Conversation object."""
        input_text = self._process_text_value(example["problem"])
        input_text = input_text.removeprefix("<image>")
        if not self._add_system_instruction:
            input_text = input_text + " " + _GEOMETRY3K_INSTRUCTION

        output_text = self._process_text_value(example["answer"])
        image = example["images"]
        if len(image) != 1:
            raise ValueError(
                f"Expected 1 image, but got {len(image)} images in example."
            )
        image = image[0]

        messages = (
            [
                Message(
                    role=Role.SYSTEM,
                    content=_GEOMETRY3K_INSTRUCTION,
                )
            ]
            if self._add_system_instruction
            else []
        ) + [
            Message(
                role=Role.USER,
                content=[
                    ContentItem(
                        type=Type.IMAGE_BINARY,
                        binary=image["bytes"],
                    ),
                    ContentItem(type=Type.TEXT, content=input_text),
                ],
            ),
            Message(role=Role.ASSISTANT, content=output_text),
        ]

        return Conversation(messages=messages)



import pytest
import torch
import torch.nn as nn
import transformers
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForVision2Seq,
)
from transformers.models.mllama.modeling_mllama import (
    MllamaCrossAttentionDecoderLayer,
    MllamaSelfAttentionDecoderLayer,
    MllamaVisionEncoderLayer,
)

from oumi.utils.torch_naming_heuristics import (
    disable_dropout,
    group_trainable_params,
    guess_transformer_layer_cls,
    resolve_transformer_layer_cls_string_as_module_set,
    simplify_transformer_layer_cls_string,
)
from tests.markers import requires_hf_token


def test_disable_dropout():
    config = transformers.GPT2Config()
    config.attn_pdrop = 0.1
    config.embd_pdrop = 0.2
    config.resid_pdrop = 0.3
    config.summary_first_dropout = 0.4
    config.vocab_size = 5
    config.initializer_range = 0.06

    disable_dropout(config)

    assert config.attn_pdrop == 0.0
    assert config.embd_pdrop == 0.0
    assert config.resid_pdrop == 0.0
    assert config.summary_first_dropout == 0.0
    assert config.vocab_size == 5
    assert config.initializer_range == 0.06


def test_group_trainable_params():
    embedding = nn.Embedding(20, 10)
    linear_bias = nn.Linear(10, 10, bias=True)
    layernorm = nn.LayerNorm(10)
    model = nn.ModuleList([embedding, linear_bias, layernorm])

    decay_params = [embedding.weight, linear_bias.weight]
    nodecay_params = [linear_bias.bias, layernorm.weight, layernorm.bias]
    expected = [
        {
            "params": decay_params,
            "weight_decay": 0.1,
        },
        {
            "params": nodecay_params,
            "weight_decay": 0.0,
        },
    ]

    assert group_trainable_params(model, 0.1) == expected


def test_guess_transformer_layer_empty_model():
    # Test with an empty model
    empty_model = nn.Module()
    with pytest.raises(ValueError, match="Unable to guess transformer layer class"):
        guess_transformer_layer_cls(empty_model)


MODEL_CONFIGS = [
    ("gpt2", "GPT2Block", AutoModelForCausalLM),
    ("facebook/opt-125m", "OPTDecoderLayer", AutoModelForCausalLM),
    ("bert-base-uncased", "BertLayer", AutoModelForCausalLM),
    ("roberta-base", "RobertaLayer", AutoModelForCausalLM),
    ("t5-small", "T5Block", AutoModel),
    (
        "HuggingFaceFW/ablation-model-fineweb-v1",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.3-70B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.2-1B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.2-3B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.1-8B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.1-70B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    (
        "meta-llama/Llama-3.1-405B-Instruct",
        "LlamaDecoderLayer",
        AutoModelForCausalLM,
    ),
    ("meta-llama/Meta-Llama-3-8B-Instruct", "LlamaDecoderLayer", AutoModelForCausalLM),
    ("meta-llama/Meta-Llama-3-70B-Instruct", "LlamaDecoderLayer", AutoModelForCausalLM),
    ("microsoft/Phi-3-mini-4k-instruct", "Phi3DecoderLayer", AutoModelForCausalLM),
    # Only available on nightly build
    # ("Qwen/Qwen2-VL-2B-Instruct", "QwenDecoderLayer", AutoModelForVision2Seq),
    ("llava-hf/llava-1.5-7b-hf", "CLIPEncoderLayer", AutoModelForVision2Seq),
    ("Salesforce/blip2-opt-2.7b", "Blip2EncoderLayer", AutoModelForVision2Seq),
    ("mistralai/Mistral-7B-v0.1", "MistralDecoderLayer", AutoModelForCausalLM),
    ("google/gemma-2-2b-it", "GemmaDecoderLayer", AutoModelForCausalLM),
    ("google/gemma-2-2b", "GemmaDecoderLayer", AutoModelForCausalLM),
]


def _load_model_architecture(model_name, builder_class):
    # Loads only the configuration to avoid downloading the full model
    # If possible, load on meta device to avoid initializing weights on CPU
    # Uses CPU for models that don't support `meta`
    try:
        with torch.device("cpu"):
            config = AutoConfig.from_pretrained(model_name, trust_remote_code=True)
            model = builder_class.from_config(config, trust_remote_code=True)
    except RuntimeError as e:
        if "Tensor.item() cannot be called on meta tensors" in str(e):
            model = builder_class.from_config(config, trust_remote_code=True)
        else:
            raise e

    return config, model


@pytest.mark.skip("Very slow test. Only run occasionally if changing that logic.")
@pytest.mark.parametrize(
    "model_name, expected_layer_name, builder_class", MODEL_CONFIGS
)
@requires_hf_token()
def test_guess_transformer_layer_cls(model_name, expected_layer_name, builder_class):
    _config, model = _load_model_architecture(model_name, builder_class)

    # Guess the transformer layer class
    layer_cls = guess_transformer_layer_cls(model)

    # Check if the guessed class name matches the expected name
    assert layer_cls.__name__ == expected_layer_name


@pytest.mark.parametrize(
    "input_name, simplified_name",
    [
        ("", ""),
        ("  \t\n", ""),
        ("Foo", "Foo"),
        (" Foo ", "Foo"),
        (" Foo, Bar ", "Foo,Bar"),
        ("zoo.Foo, Bar ", "Foo,Bar"),
        ("zoo.Foo,moo.Bar ", "Foo,Bar"),
        ("zoo.Foo,,,Zzz,moo.Bar ", "Foo,Zzz,Bar"),
    ],
)
def test_simplify_transformer_layer_cls_string(input_name: str, simplified_name: str):
    assert simplify_transformer_layer_cls_string(input_name) == simplified_name


def test_resolve_transformer_layer_cls_string_as_module_set():
    assert resolve_transformer_layer_cls_string_as_module_set("") == set()

    assert resolve_transformer_layer_cls_string_as_module_set(
        "transformers.models.mllama.modeling_mllama.MllamaCrossAttentionDecoderLayer"
    ) == set(
        {
            MllamaCrossAttentionDecoderLayer,
        }
    )

    assert resolve_transformer_layer_cls_string_as_module_set(
        "transformers.models.mllama.modeling_mllama.MllamaSelfAttentionDecoderLayer,"
        "transformers.models.mllama.modeling_mllama.MllamaCrossAttentionDecoderLayer,"
        "transformers.models.mllama.modeling_mllama.MllamaVisionEncoderLayer"
    ) == set(
        {
            MllamaSelfAttentionDecoderLayer,
            MllamaCrossAttentionDecoderLayer,
            MllamaVisionEncoderLayer,
        }
    )



import pytest
from transformers import AutoTokenizer

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import REGISTRY, RegistryType


def _get_all_sft_datasets_private_key() -> list[str]:
    """List all SFT datasets in the registry."""
    _EXCLUDED_DATASETS = set({"coco_captions", "vision_language_jsonl", "vl_sft"})

    datasets = []
    for key, value in REGISTRY.get_all(RegistryType.DATASET).items():
        if issubclass(value, BaseSftDataset) and key not in _EXCLUDED_DATASETS:
            datasets.append(key)
    return datasets


@pytest.mark.parametrize("dataset_key", _get_all_sft_datasets_private_key())
@pytest.mark.skip(
    reason="This test is very time consuming, and should be run manually."
)
def test_sft_datasets(dataset_key: str):
    dataset_cls = REGISTRY._registry[(dataset_key, RegistryType.DATASET)]
    tokenizer = AutoTokenizer.from_pretrained("microsoft/Phi-3-mini-4k-instruct")
    idx = 0

    # Dataset can successfully be loaded
    dataset = dataset_cls(tokenizer=tokenizer)
    assert dataset.raw(idx) is not None

    # Rows can successfully be pre-processed
    assert dataset.conversation(idx) is not None
    assert dataset.prompt(idx) is not None

    # Rows can successfully be used for training
    assert dataset[idx] is not None



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import Optional

from oumi.core.configs.params.synthesis_params import (
    AttributeCombination,
    DatasetSource,
    DocumentSource,
    ExampleSource,
    GeneralSynthesisParams,
    SampledAttribute,
)
from oumi.core.synthesis.dataset_ingestion import DatasetReader
from oumi.core.synthesis.document_ingestion import DocumentReader, DocumentSegmenter


class DatasetPlanner:
    """Planner for the dataset's attributes for inference."""

    def __init__(
        self,
        document_reader: Optional[DocumentReader] = None,
        dataset_reader: Optional[DatasetReader] = None,
    ):
        """Initialize the dataset planner."""
        self._document_reader = document_reader or DocumentReader()
        self._dataset_reader = dataset_reader or DatasetReader()

    def plan(
        self,
        synthesis_params: GeneralSynthesisParams,
        sample_count: int,
    ) -> list[dict]:
        """Setup the dataset's attributes for inference.

        This function will create a list of dictionaries, with each dictionary
        representing a sample of the dataset with a particular attribute value for
        each attribute.

        - Example sources are used to populate the dataset plan with a set of examples
          for specific attributes, with each example being used round-robin.
        - Document sources are used to populate the dataset plan with documents and/or
          document segments, each sample of a document source being used round-robin.
        - Dataset sources are used to populate the dataset plan with values for the
          attributes, with each sample of a dataset source being used round-robin.
        - Permutable attributes have their values sampled from a distribution.
        - Combination sampling overrides the distribution for particular attribute-value
          combinations.

        The final list of dictionaries will be used to create a dataset.

        Args:
            synthesis_params: The synthesis parameters.
            sample_count: The number of samples to plan.

        Returns:
            A list of dictionaries, each representing a sample of the dataset with
            the attribute values for each attribute.
        """
        if sample_count <= 0:
            raise ValueError("sample_count must be positive")

        example_sample_sets = self._ingest_example_sources(
            synthesis_params.input_examples
        )

        dataset_sample_sets = self._ingest_dataset_sources(synthesis_params.input_data)

        document_sample_sets = self._ingest_document_sources(
            synthesis_params.input_documents
        )

        permutable_attribute_samples = self._plan_permutable_attributes(
            synthesis_params.sampled_attributes,
            synthesis_params.combination_sampling,
            sample_count,
        )

        return self._create_dataset_plan(
            sample_count,
            permutable_attribute_samples,
            example_sample_sets,
            dataset_sample_sets,
            document_sample_sets,
        )

    def _create_dataset_plan(
        self,
        sample_count: int,
        permutable_attribute_samples: list[dict],
        example_sample_sets: list[list[dict]],
        dataset_sample_sets: list[list[dict]],
        document_sample_sets: list[list[dict]],
    ) -> list[dict]:
        """Create the final dataset plan."""
        samples = []
        for i in range(sample_count):
            sample = {}
            for example_set in example_sample_sets:
                index = i % len(example_set)
                sample.update(example_set[index])

            for dataset in dataset_sample_sets:
                index = i % len(dataset)
                sample.update(dataset[index])

            for document_set in document_sample_sets:
                index = i % len(document_set)
                sample.update(document_set[index])

            samples.append(sample)

        for i in range(len(permutable_attribute_samples)):
            samples[i].update(permutable_attribute_samples[i])

        if len(samples[-1].keys()) == 0:
            raise ValueError(
                "Empty sample created after planning, "
                "you must have at least one defined attribute for synthesis."
            )

        return samples

    def _ingest_example_sources(
        self,
        example_sources: Optional[list[ExampleSource]],
    ) -> list[list[dict]]:
        """Read examples from the example sources."""
        if example_sources is None or len(example_sources) == 0:
            return []

        results = [example_source.examples for example_source in example_sources]
        return results

    def _ingest_dataset_sources(
        self,
        dataset_sources: Optional[list[DatasetSource]],
    ) -> list[list[dict]]:
        """Read in datasets from the dataset sources."""
        if dataset_sources is None or len(dataset_sources) == 0:
            return []

        return [
            self._dataset_reader.read(dataset_source)
            for dataset_source in dataset_sources
        ]

    def _ingest_document_sources(
        self,
        document_sources: Optional[list[DocumentSource]],
    ) -> list[list[dict]]:
        """Read documents from the document sources and segment them if necessary."""
        if not document_sources:
            return []

        per_source_records = []
        for document_source in document_sources:
            records = []
            path = document_source.path
            documents = self._document_reader.read(path)
            non_empty_documents = [
                document
                for document in documents
                if document and document.strip() != ""
            ]
            if not non_empty_documents:
                raise ValueError(
                    "No non-empty documents were found in the document source, "
                    "please check the document source."
                )

            if document_source.segmentation_params is None:
                for document in non_empty_documents:
                    records.append({document_source.id: document})
            else:
                segmenter = DocumentSegmenter(document_source.segmentation_params)
                for document in non_empty_documents:
                    segments = segmenter.segment(document)
                    non_empty_segments = [
                        segment
                        for segment in segments
                        if segment and segment.strip() != ""
                    ]

                    if not non_empty_segments:
                        raise ValueError(
                            "Document segmentation returned only empty segments."
                        )

                    for segment in non_empty_segments:
                        record = {document_source.segmentation_params.id: segment}
                        if document_source.segmentation_params.keep_original_text:
                            record[document_source.id] = document
                        records.append(record)

            if not records:
                raise ValueError(
                    "No records were created from document source, check to ensure "
                    "that at least one document is not empty."
                )

            per_source_records.append(records)

        return per_source_records

    def _plan_permutable_attributes(
        self,
        permutable_attributes: Optional[list[SampledAttribute]],
        combination_sampling: Optional[list[AttributeCombination]],
        sample_count: int,
    ) -> list[dict]:
        if sample_count < 0:
            raise ValueError("Count must be positive")
        elif (
            sample_count == 0
            or permutable_attributes is None
            or len(permutable_attributes) == 0
        ):
            return []

        sampling_overrides = combination_sampling or []

        cumulative_override_probability = sum(
            [combination.sample_rate for combination in sampling_overrides]
        )

        # If cumulative probability is greater than 1, raise an error
        if cumulative_override_probability > 1.0:
            raise ValueError(
                "Cumulative probability of combination sampling must "
                "be less than or equal to 1.0."
            )

        # Generate `sample_count` permutations of the permutable attributes
        attribute_distributions = {
            perm_attr.id: perm_attr.get_value_distribution()
            for perm_attr in permutable_attributes
        }

        if cumulative_override_probability == 0.0:
            normalized_override_sample_rates = [0.0] * len(sampling_overrides)
        else:
            normalized_override_sample_rates = [
                combination.sample_rate / cumulative_override_probability
                for combination in sampling_overrides
            ]

        possible_sampling_overrides = [
            override.combination for override in sampling_overrides
        ]

        samples = []
        for _ in range(sample_count):
            sample_combination = {}

            random_number = random.random()

            # If random number < cumulative probability, sample from an override
            sampled_from_override = False
            if random_number < cumulative_override_probability:
                combination_to_sample = random.choices(
                    sampling_overrides,
                    normalized_override_sample_rates,
                    k=1,
                )[0]
                sample_combination = combination_to_sample.combination
                sampled_from_override = True

            original_sample_combination = {k: v for k, v in sample_combination.items()}
            # Sample remaining attributes
            while len(sample_combination.keys()) == 0 or (
                not sampled_from_override
                and _check_if_matches_override(
                    sample_combination,
                    possible_sampling_overrides,
                )
            ):
                sample_combination = {
                    k: v for k, v in original_sample_combination.items()
                }
                for perm_attr in permutable_attributes:
                    # If the attribute is already in the sample combination, skip it.
                    if perm_attr.id in sample_combination:
                        continue

                    value_distribution = attribute_distributions[perm_attr.id]
                    sample_combination[perm_attr.id] = random.choices(
                        list(value_distribution.keys()),
                        list(value_distribution.values()),
                        k=1,
                    )[0]

            samples.append(sample_combination)

        return samples


def _check_if_matches_override(
    sample_combination: dict,
    override_combinations: list[dict],
) -> bool:
    # For each override combination, check if it's contained in the sample
    for override_combination in override_combinations:
        # Check if all attributes in the override combination are in the sample
        all_attributes_present = all(
            attr in sample_combination for attr in override_combination
        )
        if not all_attributes_present:
            # Attributes not present, move on to next override
            continue

        # Check if the attribute values are the same
        all_values_match = all(
            sample_combination[attr] == override_combination[attr]
            for attr in override_combination
        )
        if not all_values_match:
            # Values don't match, move on to next override
            continue

        # We've got a match, resample
        return True

    return False



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oumi.core.registry import register_dataset
from oumi.datasets.vision_language.the_cauldron import TheCauldronDataset


@register_dataset("HuggingFaceM4/Docmatix")
class DocmatixDataset(TheCauldronDataset):
    """Dataset class for the `HuggingFaceM4/Docmatix` dataset.

    The dataset has the same data layout and format as `HuggingFaceM4/the_cauldron`
    (hence it's defined as a sub-class) but the underlying data is different.
    Unlike `HuggingFaceM4/the_cauldron`, the dataset contains many multi-image examples,
    and fewer subsets.

    Be aware that 'HuggingFaceM4/Docmatix' is a very large dataset (~0.5TB) that
    requires a lot of Internet bandwidth to download, and a lot of disk space to store,
    so only use it if you know what you're doing.

    Using the 'Docmatix' dataset in Oumi should become easier after streaming support
    is supported for custom Oumi datasets (OPE-1021).
    """

    default_dataset = "HuggingFaceM4/Docmatix"



import os
from pathlib import Path

import pytest

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
    RemoteParams,
)
from oumi.core.types import Conversation, Message, Role
from oumi.inference import SambanovaInferenceEngine


def load_config(config_path):
    """Load the inference configuration from a YAML file."""
    return InferenceConfig.from_yaml(config_path)


def initialize_engine(api_key, model_name):
    """Initialize the SambaNova inference engine."""
    return SambanovaInferenceEngine(
        model_params=ModelParams(model_name=model_name),
        generation_params=GenerationParams(
            batch_size=1,
        ),
        remote_params=RemoteParams(
            api_key=api_key,
        ),
    )


def create_conversation():
    """Create a conversation for inference."""
    return [
        Conversation(
            messages=[
                Message(
                    role=Role.SYSTEM,
                    content="Answer the question in a couple sentences.",
                ),
                Message(
                    role=Role.USER, content="What is the strength of SambaNova Systems?"
                ),
            ]
        ),
    ]


def perform_inference(engine, conversations, config):
    """Perform inference using the SambaNova engine."""
    try:
        generations = engine.infer(
            input=conversations,
            inference_config=config,
        )
        return generations
    except Exception as e:
        print("An error occurred during inference:", str(e))
        return None


@pytest.mark.e2e
def test_sambanova_inference():
    if "SAMBANOVA_API_KEY" not in os.environ:
        pytest.skip("SAMBANOVA_API_KEY is not set")

    # Set the path to the configuration file using pathlib
    config_path = Path(__file__).parent / "sambanova_infer_tutorial.yaml"
    print(config_path)

    # Load the configuration
    config = load_config(config_path)

    # Initialize the engine
    api_key = os.getenv("SAMBANOVA_API_KEY")
    model_name = "Meta-Llama-3.1-405B-Instruct"
    engine = initialize_engine(api_key, model_name)

    # Create the conversation
    conversations = create_conversation()

    # Perform inference
    generations = perform_inference(engine, conversations, config)

    # Print the results
    if generations:
        print(generations)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("CohereForAI/aya_dataset")
class AyaDataset(BaseSftDataset):
    """Dataset class for the CohereForAI/aya_dataset dataset."""

    default_dataset = "CohereForAI/aya_dataset"

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Transform a dataset example into a Conversation object."""
        question: str = example.get("inputs", None) or ""
        answer: str = example.get("targets", None) or ""

        messages = [
            Message(role=Role.USER, content=question),
            Message(role=Role.ASSISTANT, content=answer),
        ]

        return Conversation(messages=messages)



import base64
from typing import Any, Final, Optional

import numpy as np
import PIL.Image
import pytest
import torch
import transformers

from oumi.builders import build_chat_template, build_processor, build_tokenizer
from oumi.core.configs import ModelParams
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types.conversation import ContentItem, Message, Role, Type

_LLAVA_SYSTEM_PROMPT: Final[str] = (
    "A chat between a curious user and an artificial "
    "intelligence assistant. "
    "The assistant gives helpful, detailed, and "
    "polite answers to the user's questions."
)
_IMAGE_TOKEN: Final[str] = "<image>"
_IMAGE_TOKEN_ID: Final[int] = 32000

_SMALL_B64_IMAGE: Final[str] = (
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
)


@pytest.mark.parametrize(
    "trust_remote_code",
    [
        False,
        True,
    ],
)
def test_build_processor_empty_name(trust_remote_code, mock_tokenizer):
    with pytest.raises(ValueError, match="Empty model name"):
        build_processor("", mock_tokenizer, trust_remote_code=trust_remote_code)


@pytest.mark.parametrize(
    "processor_kwargs",
    [
        None,
        {},
    ],
)
def test_build_processor_basic_gpt2_success(
    processor_kwargs: Optional[dict[str, Any]], mock_tokenizer
):
    test_chat_template: Final[str] = build_chat_template(template_name="default")

    model_params = ModelParams(model_name="openai-community/gpt2")
    processor = build_processor(
        model_params.model_name,
        mock_tokenizer,
        trust_remote_code=False,
        processor_kwargs=processor_kwargs,
    )
    assert callable(processor)

    assert id(mock_tokenizer) == id(processor.tokenizer)
    processor.tokenizer = mock_tokenizer
    assert id(mock_tokenizer) == id(processor.tokenizer)
    assert processor.chat_template == test_chat_template

    processor.chat_template = test_chat_template + " "
    assert processor.chat_template == test_chat_template + " "
    processor.chat_template = test_chat_template
    assert processor.chat_template == test_chat_template

    assert processor.image_processor is None
    assert processor.image_token is None
    assert processor.image_token_id is None

    result = processor(text=["hello world"], padding=False)
    assert isinstance(result, transformers.BatchEncoding)
    assert len(result) == 2

    assert "input_ids" in result
    input_ids = result["input_ids"]
    assert isinstance(input_ids, torch.Tensor)
    assert input_ids.shape == (1, 2)
    assert np.all(input_ids.numpy() == np.array([[31373, 995]]))

    assert "attention_mask" in result
    attention_mask = result["attention_mask"]
    assert isinstance(attention_mask, torch.Tensor)
    assert attention_mask.shape == (1, 2)
    assert np.all(attention_mask.numpy() == np.array([[1, 1]]))

    prompt = processor.apply_chat_template([Message(role=Role.USER, content="FooBazz")])
    assert isinstance(prompt, str)
    assert "FooBazz" in prompt
    assert prompt == "USER: FooBazz"

    prompt = processor.apply_chat_template(
        [
            Message(role=Role.USER, content="Hello"),
            Message(role=Role.ASSISTANT, content="How can I help?"),
            Message(role=Role.USER, content="Hmm"),
        ],
        add_generation_prompt=True,
    )
    assert isinstance(prompt, str)
    assert prompt == "USER: Hello\nASSISTANT: How can I help?\nUSER: Hmm\nASSISTANT: "

    with pytest.raises(ValueError, match="Conversation includes non-text messages"):
        processor.apply_chat_template(
            [
                Message(
                    role=Role.USER,
                    content=[
                        ContentItem(type=Type.TEXT, content="Hello"),
                        ContentItem(
                            type=Type.IMAGE_BINARY,
                            binary=base64.b64decode(_SMALL_B64_IMAGE),
                        ),
                    ],
                ),
                Message(role=Role.ASSISTANT, content="How can I help?"),
                Message(role=Role.USER, content="Hmm"),
            ]
        )


def test_build_processor_basic_multimodal_success():
    default_chat_template: Final[str] = build_chat_template(template_name="default")
    llava_chat_template: Final[str] = build_chat_template(template_name="llava")

    model_params = ModelParams(
        model_name="llava-hf/llava-1.5-7b-hf", chat_template="default"
    )
    tokenizer: BaseTokenizer = build_tokenizer(model_params)
    processor: BaseProcessor = build_processor(
        model_params.model_name, tokenizer, trust_remote_code=False
    )
    assert callable(processor)

    assert id(tokenizer) == id(processor.tokenizer)
    processor.tokenizer = tokenizer
    assert id(tokenizer) == id(processor.tokenizer)
    assert processor.chat_template
    assert processor.chat_template == default_chat_template
    processor.chat_template = llava_chat_template
    assert processor.chat_template == llava_chat_template

    assert processor.image_processor is not None
    assert processor.image_token == _IMAGE_TOKEN
    assert processor.image_token_id == _IMAGE_TOKEN_ID

    result = processor(text=["hello world"], padding=False)
    assert isinstance(result, transformers.BatchEncoding)
    assert len(result) == 2

    assert "input_ids" in result
    input_ids = result["input_ids"]
    assert isinstance(input_ids, torch.Tensor)
    assert input_ids.shape == (1, 3)
    assert np.all(input_ids.numpy() == np.array([[1, 22172, 3186]]))

    assert "attention_mask" in result
    attention_mask = result["attention_mask"]
    assert isinstance(attention_mask, torch.Tensor)
    assert attention_mask.shape == (1, 3)
    assert np.all(attention_mask.numpy() == np.array([[1, 1, 1]]))

    prompt = processor.apply_chat_template([Message(role=Role.USER, content="FooBazz")])
    assert isinstance(prompt, str)
    assert "FooBazz" in prompt
    assert prompt == _LLAVA_SYSTEM_PROMPT + " USER: FooBazz "

    prompt = processor.apply_chat_template(
        [
            Message(role=Role.USER, content="Hello"),
            Message(role=Role.ASSISTANT, content="How can I help?"),
            Message(role=Role.USER, content="Hmm"),
        ],
        add_generation_prompt=True,
    )
    assert isinstance(prompt, str)
    assert prompt == (
        _LLAVA_SYSTEM_PROMPT
        + " USER: Hello ASSISTANT: How can I help? </s>USER: Hmm ASSISTANT: "
    )

    test_image = PIL.Image.new(mode="RGB", size=(512, 256))
    result = processor(
        text=[prompt], images=[test_image], padding=True, return_tensors="pt"
    )
    assert isinstance(result, transformers.BatchEncoding)
    assert sorted(list(result.keys())) == [
        "attention_mask",
        "input_ids",
        "pixel_values",
    ]
    attention_mask = result["attention_mask"]
    assert isinstance(attention_mask, torch.Tensor)
    assert attention_mask.shape == (1, 57)

    input_ids = result["input_ids"]
    assert isinstance(input_ids, torch.Tensor)
    assert input_ids.shape == (1, 57)

    pixel_values = result["pixel_values"]
    assert isinstance(pixel_values, torch.Tensor)
    assert pixel_values.shape == (1, 3, 336, 336)

    image_proc_result = processor.image_processor(
        images=[test_image], return_tensors="pt"
    )
    assert isinstance(image_proc_result, transformers.BatchFeature)
    assert sorted(list(image_proc_result.keys())) == [
        "pixel_values",
    ]
    image_proc_pixel_values = result["pixel_values"]
    assert isinstance(image_proc_pixel_values, torch.Tensor)
    assert image_proc_pixel_values.shape == (1, 3, 336, 336)

    assert np.all(image_proc_pixel_values.numpy() == pixel_values.numpy())

    # Multiple prompts, Multiple images (different counts).
    result = processor(
        text=[prompt, prompt, prompt],
        images=[test_image, test_image],
        padding=True,
        return_tensors="pt",
    )
    assert isinstance(result, transformers.BatchEncoding)
    assert sorted(list(result.keys())) == [
        "attention_mask",
        "input_ids",
        "pixel_values",
    ]
    attention_mask = result["attention_mask"]
    assert isinstance(attention_mask, torch.Tensor)
    assert attention_mask.shape == (3, 57)

    input_ids = result["input_ids"]
    assert isinstance(input_ids, torch.Tensor)
    assert input_ids.shape == (3, 57)

    pixel_values = result["pixel_values"]
    assert isinstance(pixel_values, torch.Tensor)
    assert pixel_values.shape == (2, 3, 336, 336)

    image_proc_result = processor.image_processor(
        images=[test_image, test_image], return_tensors="pt"
    )
    assert isinstance(image_proc_result, transformers.BatchFeature)
    assert sorted(list(image_proc_result.keys())) == [
        "pixel_values",
    ]
    image_proc_pixel_values = image_proc_result["pixel_values"]
    assert isinstance(image_proc_pixel_values, torch.Tensor)
    assert image_proc_pixel_values.shape == (2, 3, 336, 336)

    assert np.all(image_proc_pixel_values.numpy() == pixel_values.numpy())



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import AsyncMock

import aiohttp
import pytest

from oumi.utils.http import (
    get_failure_reason_from_response,
    is_non_retriable_status_code,
)


@pytest.mark.parametrize(
    "status_code,expected",
    [
        (400, True),  # Bad Request
        (401, True),  # Unauthorized
        (403, True),  # Forbidden
        (404, True),  # Not Found
        (422, True),  # Unprocessable Entity
        (500, False),  # Server Error
        (502, False),  # Bad Gateway
        (503, False),  # Service Unavailable
        (429, False),  # Too Many Requests
    ],
)
def test_is_non_retryable_status_code(status_code: int, expected: bool):
    """Test identification of non-retryable status codes."""
    assert is_non_retriable_status_code(status_code) == expected


@pytest.mark.asyncio
async def test_get_failure_reason_from_response_with_json_response():
    """Test handling of non-retryable errors with JSON response."""
    mock_response = AsyncMock(spec=aiohttp.ClientResponse)
    mock_response.status = 400
    mock_response.json.return_value = {"error": {"message": "Invalid request"}}

    result = await get_failure_reason_from_response(mock_response)
    assert result == "Invalid request"


@pytest.mark.asyncio
async def test_get_failure_reason_from_response_with_list_response():
    """Test handling of non-retryable errors with list response."""
    mock_response = AsyncMock(spec=aiohttp.ClientResponse)
    mock_response.status = 400
    mock_response.json.return_value = [{"error": {"message": "Invalid request"}}]

    result = await get_failure_reason_from_response(mock_response)
    assert result == "Invalid request"


@pytest.mark.asyncio
async def test_get_failure_reason_from_response_with_empty_response():
    """Test handling of non-retryable errors with empty response."""
    mock_response = AsyncMock(spec=aiohttp.ClientResponse)
    mock_response.status = 400
    mock_response.json.return_value = {}

    result = await get_failure_reason_from_response(mock_response)
    assert result == "HTTP 400"


@pytest.mark.asyncio
async def test_get_failure_reason_from_response_with_json_error():
    """Test handling of non-retryable errors when JSON parsing fails."""
    mock_response = AsyncMock(spec=aiohttp.ClientResponse)
    mock_response.status = 400
    mock_response.json.side_effect = Exception("JSON decode error")

    result = await get_failure_reason_from_response(mock_response)
    assert result == "HTTP 400"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

from datasets import Dataset, DatasetDict, load_dataset

# MMLU prompts are classified into 57 subjects and 'all' (which contains all subjects)
SUBJECTS = [
    "abstract_algebra",
    "all",
    "anatomy",
    "astronomy",
    "business_ethics",
    "clinical_knowledge",
    "college_biology",
    "college_chemistry",
    "college_computer_science",
    "college_mathematics",
    "college_medicine",
    "college_physics",
    "computer_security",
    "conceptual_physics",
    "econometrics",
    "electrical_engineering",
    "elementary_mathematics",
    "formal_logic",
    "global_facts",
    "high_school_biology",
    "high_school_chemistry",
    "high_school_computer_science",
    "high_school_european_history",
    "high_school_geography",
    "high_school_government_and_politics",
    "high_school_macroeconomics",
    "high_school_mathematics",
    "high_school_microeconomics",
    "high_school_physics",
    "high_school_psychology",
    "high_school_statistics",
    "high_school_us_history",
    "high_school_world_history",
    "human_aging",
    "human_sexuality",
    "international_law",
    "jurisprudence",
    "logical_fallacies",
    "machine_learning",
    "management",
    "marketing",
    "medical_genetics",
    "miscellaneous",
    "moral_disputes",
    "moral_scenarios",
    "nutrition",
    "philosophy",
    "prehistory",
    "professional_accounting",
    "professional_law",
    "professional_medicine",
    "professional_psychology",
    "public_relations",
    "security_studies",
    "sociology",
    "us_foreign_policy",
    "virology",
    "world_religions",
]
SPLITS = [
    "dev",  # For few-shot development; 5 questions per subject (285 questions)
    "validation"  # For selecting hyperparameters (1,531 questions)
    "test",  # For testing purposes (14,042 questions)
]
DEFAULT_NUM_SHOTS = 0  # Values: 0-5; 0 is consistent with LM Evaluation Harness.


# FIXME: Inherit from `BaseMapDataset`.
class MmluDataset:
    # MMLU questions always have 4 possible answers, which are labelled A, B, C, D.
    answer_tokens = ["A", "B", "C", "D"]

    # Static and Class methods for formatting prompts.
    @staticmethod
    def format_subject(subject: str) -> str:
        """Formats the subject of the prompt."""
        return subject.replace("_", " ")

    @classmethod
    def format_example(
        cls, example: dict[str, Any], include_answer: bool = True
    ) -> str:
        """Formats an MMLU example."""
        prompt = example["question"]
        for index, choice in enumerate(example["choices"]):
            prompt += f"\n{cls.answer_tokens[index]}. {choice}"
        prompt += "\nAnswer:"
        if include_answer:
            correct_answer_index = example["answer"]
            prompt += f" {cls.answer_tokens[correct_answer_index]}\n\n"
        return prompt

    @classmethod
    def few_shots(cls, dev_data: Dataset, num_shots: int) -> str:
        """Returns `num_shots` formatted shots from the provided `dev_data`."""
        if not num_shots:
            return ""
        shots: Dataset = dev_data.select(range(num_shots))
        return "".join(
            cls.format_example(example, include_answer=True)  # type: ignore
            for example in shots  # type: ignore
        )

    def __init__(self, subject: str = "all", num_shots: Optional[int] = None):
        """Initializes the class MmluDataset."""
        if subject not in SUBJECTS:
            raise ValueError(f"MMLU: unknown subject `{subject}`")
        self._dataset_dict: DatasetDict = load_dataset("cais/mmlu", subject)  # type: ignore
        self._few_shot_dict: dict[str, str] = dict()
        self._num_shots = num_shots if (num_shots is not None) else DEFAULT_NUM_SHOTS

    # Instance methods (private).
    def _prompt_template(self, example: dict[str, Any]) -> str:
        """Generates the prompt template for evaluations.

        This template is the "original" MMLU implementation by github.com/ollmer.
        For details and undertanding the different options to evaluate with MMLU, see:
        https://huggingface.co/blog/open-llm-leaderboard-mmlu.

        After appying the template, the prompt format will look as follows:
        --------------------------------------------------------------------------------
        The following are multiple choice questions (with answers) about <Subject>.

        <Question-1 from `dev` dataset>
        A. <answer 1.1>
        B. <answer 1.2>
        C. <answer 1.3>
        D. <answer 1.4>
        Answer: B

        <Question-2 from `dev` dataset>
        A. <answer 2.1>
        B. <answer 2.2>
        C. <answer 2.3>
        D. <answer 2.4>
        Answer: A

        [...]

        <Question-5 from `dev` dataset>
        A. <answer 5.1>
        B. <answer 5.2>
        C. <answer 5.3>
        D. <answer 5.4>
        Answer: C

        <Question from `example`>
        A. <answer t.1>
        B. <answer t.2>
        C. <answer t.3>
        D. <answer t.4>
        Answer:
        --------------------------------------------------------------------------------
        """
        subject = example["subject"]
        if subject not in self._few_shot_dict:
            self._update_few_shot_dict(subject)  # Lazy initialization.

        prompt = (
            f"The following are multiple choice questions (with answers) "
            f"about {MmluDataset.format_subject(subject)}.\n\n"
        )
        prompt += self._few_shot_dict[subject]
        prompt += MmluDataset.format_example(example, include_answer=False)
        return prompt

    def _update_few_shot_dict(self, subject: str):
        """Adds few-shot examples for `subject` in the relevant dictionary."""
        assert subject in SUBJECTS
        dataset_dict: DatasetDict = load_dataset("cais/mmlu", subject)  # type: ignore
        dev_dataset: Dataset = dataset_dict["dev"]
        few_shots = MmluDataset.few_shots(dev_dataset, self._num_shots)
        self._few_shot_dict[subject] = few_shots

    def _get_dataset(self, split: str, num_entries: Optional[int] = None) -> Dataset:
        dataset: Dataset = self._dataset_dict[split]
        if num_entries:
            dataset = dataset.select(range(num_entries))
        return dataset

    def _get_formatted_dataset(
        self, split: str, num_entries: Optional[int] = None
    ) -> list[str]:
        dataset: Dataset = self._get_dataset(split=split, num_entries=num_entries)
        dataset_formatted: list[str] = list(map(self._prompt_template, dataset))  # type: ignore
        return dataset_formatted

    def _get_labels(self, split: str, num_entries: Optional[int] = None) -> list[int]:
        dataset: Dataset = self._dataset_dict[split]
        if num_entries:
            dataset = dataset.select(range(num_entries))
        return [example["answer"] for example in dataset]  # type: ignore

    # Instance methods (global).
    # All these will potentially be required by the base `BaseMapDataset`.
    def get_test_split(self, num_entries: Optional[int] = None) -> list[str]:
        """Returns the test split of this dataset."""
        return self._get_formatted_dataset(split="test", num_entries=num_entries)

    def get_validation_split(self, num_entries: Optional[int] = None) -> list[str]:
        """Returns the validation split of this dataset."""
        return self._get_formatted_dataset(split="validation", num_entries=num_entries)

    def get_test_labels(self, num_entries: Optional[int] = None) -> list[int]:
        """Returns the labels of the test dataset."""
        return self._get_labels(split="test", num_entries=num_entries)

    def get_validation_labels(self, num_entries: Optional[int] = None) -> list[int]:
        """Returns the labels of the validation dataset."""
        return self._get_labels(split="validation", num_entries=num_entries)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Launcher module for the Oumi (Open Universal Machine Intelligence) library.

This module provides base classes for cloud and cluster management in
the Oumi framework.

These classes serve as foundations for implementing cloud-specific and cluster-specific
launchers for running machine learning jobs.
"""

from oumi.core.launcher.base_cloud import BaseCloud
from oumi.core.launcher.base_cluster import BaseCluster, JobState, JobStatus

__all__ = [
    "BaseCloud",
    "BaseCluster",
    "JobState",
    "JobStatus",
]



from oumi.core.configs import EvaluationConfig, InferenceConfig, TrainingConfig
from oumi.evaluate import evaluate
from oumi.infer import infer
from oumi.train import train
from oumi.utils.torch_utils import device_cleanup


def main() -> None:
    """Run Llama 1B train/eval/infer."""
    model_output_dir = "output/llama1b_e2e"
    device_cleanup()
    train_config: TrainingConfig = TrainingConfig.from_yaml(
        "configs/recipes/llama3_2/sft/1b_full/train.yaml"
    )
    train_config.training.enable_wandb = False
    train_config.training.max_steps = 100
    train_config.training.output_dir = model_output_dir
    train_config.finalize_and_validate()
    train(train_config)

    device_cleanup()
    eval_config: EvaluationConfig = EvaluationConfig.from_yaml(
        "configs/recipes/llama3_2/evaluation/1b_eval.yaml"
    )
    eval_config.model.model_name = model_output_dir
    eval_config.finalize_and_validate()
    evaluate(eval_config)

    device_cleanup()
    infer_config: InferenceConfig = InferenceConfig.from_yaml(
        "configs/recipes/llama3_2/inference/1b_infer.yaml"
    )
    infer_config.model.model_name = model_output_dir
    infer_config.finalize_and_validate()
    model_responses = infer(
        config=infer_config,
        inputs=[
            "Foo",
            "Bar",
        ],
    )
    print(model_responses)

    device_cleanup()


if __name__ == "__main__":
    main()
